"""
Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE

Contains large portions of:
https://github.com/Rapptz/discord.py/tree/master/discord/gateway.py
The appropriate license is in LICENSE
"""

from __future__ import annotations

import asyncio
import base64
import concurrent.futures
import logging
import random
import sys
import threading
import time
import traceback
from contextlib import asynccontextmanager
from dataclasses import dataclass
from datetime import datetime, timedelta
from functools import partial
from gzip import FCOMMENT, FEXTRA, FHCRC, FNAME
from ipaddress import IPv4Address
from operator import attrgetter
from types import CoroutineType
from typing import TYPE_CHECKING, Any, Final, Generic, Self, TypeAlias, overload
from zlib import MAX_WBITS, decompress

import aiohttp
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from typing_extensions import TypeVar

from . import utils
from ._const import CLEAR_PROTO_BIT, DEFAULT_CMS, IS_PROTO, READ_U32, SET_PROTO_BIT, timeout
from .enums import *
from .errors import AuthenticatorError, HTTPException, NoCMsFound, WSException
from .id import parse_id64
from .models import return_true
from .protobufs import (
    EMsg,
    GCMessage,
    GCProtobufMessage,
    Message,
    ProtobufMessage,
    UnifiedMessage,
    auth,
    chat,
    client_server,
    client_server_2,
    friends,
    login,
)
from .types.id import ID64, AppID
from .user import AnonymousClientUser, ClientUser

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Iterable

    from .client import Client
    from .enums import UIMode
    from .protobufs.base import CMsgMulti
    from .state import ConnectionState
    from .types.http import Coro, IPAdress


__all__ = (
    "ConnectionClosed",
    "SteamWebSocket",
    "CMServer",
    "Msgs",
    "RAISED_EXCEPTIONS",
)

log = logging.getLogger(__name__)
ProtoMsgs: TypeAlias = ProtobufMessage | Message
GCMsgs: TypeAlias = GCProtobufMessage | GCMessage

GCMsgsT = TypeVar("GCMsgsT", bound=GCMsgs, default=GCMsgs)
ProtoMsgsT = TypeVar("ProtoMsgsT", bound=ProtoMsgs, default=ProtoMsgs)

Msgs: TypeAlias = ProtoMsgs | GCMsgs
MsgsT = TypeVar("MsgsT", bound=Msgs, default=Msgs)

ProtoMsgT = TypeVar("ProtoMsgT", bound=ProtobufMessage, default=ProtobufMessage)
UnifiedMsgT = TypeVar("UnifiedMsgT", bound=UnifiedMessage, default=UnifiedMessage)
MsgT = TypeVar("MsgT", bound=Message, default=Message)
GCMsgT = TypeVar("GCMsgT", bound=GCMessage, default=GCMessage)
GCMsgProtoT = TypeVar("GCMsgProtoT", bound=GCProtobufMessage, default=GCProtobufMessage)

PROTOCOL_VERSION: Final = 65580


@dataclass(slots=True)
class EventListener(Generic[MsgsT]):
    msg: IntEnum | None
    check: Callable[[MsgsT], bool]
    future: asyncio.Future[MsgsT]

    if not TYPE_CHECKING:
        __class_getitem__ = classmethod(lambda cls, params: cls)


@dataclass(slots=True)
class GCEventListener(EventListener[GCMsgsT]):
    app_id: AppID


@dataclass(slots=True)
class CMServer:
    _state: ConnectionState
    url: str
    weighted_load: float

    def connect(self) -> Coro[aiohttp.ClientWebSocketResponse]:
        return self._state.http.connect_to_cm(self.url)

    async def ping(self) -> float:
        try:
            async with timeout(5), self._state.http._session.get(f"https://{self.url}/cmping/") as resp:
                if resp.status != 200:
                    raise KeyError
                return int(resp.headers["X-Steam-CMLoad"])
        except (KeyError, asyncio.TimeoutError, aiohttp.ClientError):
            return float("inf")


async def fetch_cm_list(state: ConnectionState, cell_id: int = 0) -> AsyncGenerator[CMServer, None]:
    if state._connected_cm is not None:
        yield state._connected_cm
    log.debug("Attempting to fetch servers from the WebAPI")
    state.cell_id = cell_id
    try:
        data = await state.http.get_cm_list(cell_id)
    except Exception:
        servers = [CMServer(state, url=cm_url, weighted_load=0) for cm_url in DEFAULT_CMS]
        random.shuffle(servers)
        log.debug("Error occurred when fetching CM server list, falling back to internal list", exc_info=True)
        for cm in servers:
            yield cm
        return

    if not data["success"]:
        servers = [CMServer(state, url=cm_url, weighted_load=0) for cm_url in DEFAULT_CMS]
        random.shuffle(servers)
        log.debug("Error occurred when fetching CM server list, falling back to internal list", exc_info=True)
        for cm in servers:
            yield cm
        return

    hosts = data["serverlist"]
    log.debug("Received %d servers from WebAPI", len(hosts))
    for cm in sorted(
        (CMServer(state, url=server["endpoint"], weighted_load=server["wtd_load"]) for server in hosts),
        key=attrgetter("weighted_load"),
    ):  # they should already be sorted but oh well
        yield cm


def unpack_multi(msg: CMsgMulti) -> bytes | None:
    data = msg.message_body
    if data[:2] != b"\037\213":
        return log.info("Received a file that's not GZipped")

    position = 10

    if flag := data[3]:  # this isn't ever hit, might as well save a few nanos
        if flag & FEXTRA:
            extra_len = int.from_bytes(data[position : position + 2], "little")
            position += 2 + extra_len
        if flag & FNAME:
            while data[position:]:
                position += 1
        if flag & FCOMMENT:
            while data[position:]:
                position += 1
        if flag & FHCRC:
            position += 2

    decompressed = decompress(data[position:], wbits=-MAX_WBITS)

    if len(decompressed) != msg.size_unzipped:
        return log.info("Unzipped size mismatch for multi payload %r, discarding", msg)

    return decompressed


class ConnectionClosed(Exception):
    def __init__(self, cm: CMServer):
        self.cm = cm
        super().__init__(f"Connection to {self.cm.url}, has closed.")


class WebSocketClosure(Exception):
    """An exception to make up for the fact that aiohttp doesn't signal closure."""


RAISED_EXCEPTIONS: Final = (
    OSError,
    ConnectionClosed,
    aiohttp.ClientError,
    asyncio.TimeoutError,
    HTTPException,
)


class KeepAliveHandler(threading.Thread):
    def __init__(self, ws: SteamWebSocket, interval: int, loop: asyncio.AbstractEventLoop):
        super().__init__()
        self.ws = ws
        self.loop = loop
        self.interval = interval
        self._main_thread_id = self.ws.thread_id
        self.heartbeat = login.CMsgClientHeartBeat(send_reply=True)
        self.msg = "Keeping websocket alive with heartbeat %s."
        self.block_msg = "Heartbeat blocked for more than {total} seconds."
        self.behind_msg = "Can't keep up, websocket is {total:.1f}s behind."
        self._stop_ev = threading.Event()
        self._last_recv = float("-inf")
        self._last_ack = time.perf_counter()
        self._last_send = time.perf_counter()
        self.latency = float("inf")

    def run(self) -> None:
        while not self._stop_ev.wait(self.interval):
            if self._last_recv + 60 < time.perf_counter():
                log.warning("CM %r has stopped responding to the gateway. Closing and restarting.", self.ws.cm)
                coro = self.ws.close(4000)
                f = asyncio.run_coroutine_threadsafe(coro, loop=self.loop)

                try:
                    f.result()
                except Exception:
                    log.exception("An error occurred while stopping the gateway. Ignoring.")
                finally:
                    return self.stop()

            log.debug(self.msg, self.heartbeat)
            if self.loop.is_closed():
                return self.stop()

            coro = self.ws.send_proto(self.heartbeat)
            f = asyncio.run_coroutine_threadsafe(coro, loop=self.loop)
            # block until sending is complete
            total = 0
            try:
                while True:
                    try:
                        f.result(timeout=10)
                        break
                    except concurrent.futures.TimeoutError:
                        total += 10
                        try:
                            frame = sys._current_frames()[self._main_thread_id]
                        except KeyError:
                            msg = self.block_msg
                        else:
                            stack = "".join(traceback.format_stack(frame))
                            msg = f"{self.block_msg}\nLoop thread traceback (most recent call last):\n{stack}"
                        log.warning(msg.format(total=total))

            except Exception:
                self.stop()
            else:
                self._last_send = time.perf_counter()

    def stop(self) -> None:
        self._stop_ev.set()

    def tick(self) -> None:
        self._last_recv = time.perf_counter()

    def ack(self) -> None:
        ack_time = time.perf_counter()
        self._last_ack = ack_time
        self.latency = ack_time - self._last_send
        if self.latency > 10:
            log.warning(self.behind_msg.format(total=self.latency))


class SteamWebSocket:
    def __init__(
        self,
        state: ConnectionState,
        socket: aiohttp.ClientWebSocketResponse,
        cm_list: AsyncGenerator[CMServer, None],
        cm: CMServer,
    ):
        self.socket = socket

        # state stuff
        self._state = state
        self.cm_list = cm_list
        self.cm = cm
        self.tg = asyncio.TaskGroup()
        # the keep alive
        self._keep_alive: KeepAliveHandler
        self._dispatch = state.dispatch
        self.thread_id = threading.get_ident()

        # ws related stuff
        self.listeners: list[EventListener[Any]] = []
        self.gc_listeners: list[GCEventListener[Any]] = []
        self.closed = False

        self.session_id = 0
        self.id64 = ID64(0)
        self._current_job_id = 0
        self._gc_current_job_id = 0
        self.server_offset = timedelta()
        self.refresh_token: str
        self._access_token: str
        self.client_id: int

        self.public_ip: IPAdress
        self.connect_time: datetime

    async def __aenter__(self) -> Self:
        await self.tg.__aenter__()
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.tg.__aexit__(*args)

    @property
    def latency(self) -> float:
        """Measures latency between a heartbeat send and the heartbeat interval in seconds."""
        return self._keep_alive.latency

    @overload
    def wait_for(
        self, /, *, emsg: EMsg | None, check: Callable[[ProtoMsgsT], bool] = return_true
    ) -> asyncio.Future[ProtoMsgsT]: ...

    @overload
    def wait_for(self, msg: type[MsgT], /, *, check: Callable[[MsgT], bool] = return_true) -> asyncio.Future[MsgT]: ...

    @overload
    def wait_for(
        self, msg: type[ProtoMsgT], /, *, check: Callable[[ProtoMsgT], bool] = return_true
    ) -> asyncio.Future[ProtoMsgT]: ...

    def wait_for(
        self,
        msg: type[ProtoMsgsT] | None = None,
        /,
        *,
        emsg: EMsg | None = None,
        check: Callable[[ProtoMsgsT], bool] = return_true,
    ) -> asyncio.Future[ProtoMsgsT]:
        future: asyncio.Future[ProtoMsgsT] = asyncio.get_running_loop().create_future()
        entry = EventListener(msg=msg.MSG if msg else emsg, check=check, future=future)
        self.listeners.append(entry)
        return future

    @overload
    def gc_wait_for(
        self, /, *, emsg: IntEnum | None, app_id: AppID | None = None, check: Callable[[GCMsgsT], bool] = return_true
    ) -> asyncio.Future[GCMsgsT]: ...

    @overload
    def gc_wait_for(
        self, msg: type[GCMsgT], /, *, check: Callable[[GCMsgT], bool] = return_true
    ) -> asyncio.Future[GCMsgT]: ...

    @overload
    def gc_wait_for(
        self, msg: type[GCMsgProtoT], /, *, check: Callable[[GCMsgProtoT], bool] = return_true
    ) -> asyncio.Future[GCMsgProtoT]: ...

    def gc_wait_for(
        self,
        msg: type[GCMsgsT] | None = None,
        /,
        *,
        emsg: IntEnum | None = None,
        app_id: AppID | None = None,
        check: Callable[[GCMsgsT], bool] = return_true,
    ) -> asyncio.Future[GCMsgsT]:
        from ._gc import APP

        future: asyncio.Future[GCMsgsT] = asyncio.get_running_loop().create_future()
        entry: GCEventListener[GCMsgsT] = GCEventListener(
            msg=msg.MSG if msg else emsg,
            check=check,
            future=future,
            app_id=msg.APP_ID if msg else app_id or APP.get().id,
        )
        self.gc_listeners.append(entry)
        return future

    @asynccontextmanager
    async def poll(self) -> AsyncGenerator[None, None]:
        timeout = asyncio.Timeout(None)
        async with asyncio.TaskGroup() as tg:

            async def inner_poll():
                try:
                    async with timeout:
                        while True:
                            await self.poll_event()
                except TimeoutError:
                    pass

            tg.create_task(inner_poll())
            yield
            timeout.reschedule(-1)

    def dispatch_ready(self):
        state = self._state

        async def inner():
            if state.intents & Intents.ChatGroups > 0:
                await state.handled_chat_groups.wait()  # ensure group cache is ready
            if state.intents & Intents.ChatGroups > 0:
                # due to a steam limitation we can't get these reliably on reconnect?  TODO check?
                await state.handled_friends.wait()  # ensure friend cache is ready
            await state.handled_emoticons.wait()  # ensure emoticon cache is ready
            await state.handled_licenses.wait()  # ensure licenses are ready
            await state.handled_wallet.wait()  # ensure wallet is ready

            await state.client._handle_ready()

        self.tg.create_task(inner())

    @classmethod
    async def from_client(
        cls, client: Client, /, refresh_token: str | None = None, cm_list: AsyncGenerator[CMServer, None] | None = None
    ) -> SteamWebSocket:
        state = client._state
        cm_list = cm_list or fetch_cm_list(state)

        async for cm in cm_list:
            log.info("Attempting to create a websocket connection to %s (load: %f)", cm.url, cm.weighted_load)
            try:
                socket = await cm.connect()
            except aiohttp.ClientError:
                continue

            log.debug("Connected to %s", cm.url)

            self = cls(state, socket, cm_list, cm)
            old_tg = self.tg
            self.tg = client._tg
            client.ws = self
            self._dispatch("connect")

            async with self.poll():
                await self.send_proto(login.CMsgClientHello(PROTOCOL_VERSION))

                self.refresh_token = refresh_token or await self.fetch_refresh_token()
                self.id64 = parse_id64(utils.decode_jwt(self.refresh_token)["sub"])

                self.dispatch_ready()
                msg: login.CMsgClientLogonResponse = await self.send_proto_and_wait(
                    login.CMsgClientLogon(
                        protocol_version=PROTOCOL_VERSION,
                        client_package_version=1561159470,
                        client_os_type=16,
                        client_language=state.language.api_name,
                        supports_rate_limit_response=True,
                        chat_mode=2,
                        access_token=self.refresh_token,  # lol
                    ),
                    check=lambda msg: isinstance(msg, login.CMsgClientLogonResponse),
                )
                if msg.result != Result.OK:
                    log.debug("Failed to login with result: %r", msg.result)
                    await self._state.handle_close()

                self.public_ip = IPv4Address(msg.public_ip.v4)
                self.connect_time = utils.DateTime.now()
                self.session_id = msg.header.session_id

                us = await anext(self.fetch_users((self.id64,)))
                client.http.user = ClientUser(state, us)
                if hasattr(self._state, "_original_client_user_msg"):
                    self._state._original_client_user_msg = us  # type: ignore
                state._users[client.user.id] = client.user  # type: ignore
                self._state.cell_id = msg.cell_id

                self._keep_alive = KeepAliveHandler(
                    ws=self, interval=msg.heartbeat_seconds, loop=asyncio.get_running_loop()
                )
                self._keep_alive.start()
                log.debug("Heartbeat started.")

                state.login_complete.set()

                await self.send_um(chat.GetMyChatRoomGroupsRequest())
                await self.send_proto(friends.CMsgClientGetEmoticonList())
                await self._state.fetch_notifications()
                await self.send_proto(
                    login.CMsgClientServerTimestampRequest(client_request_timestamp=int(time.time() * 1000))
                )
                await self.change_presence(
                    apps=self._state._apps,
                    state=self._state._state,
                    flags=self._state._flags,
                    force_kick=self._state._force_kick,
                )

                self._dispatch("login")
                log.debug("Logon completed")
                self.tg = old_tg

                return self
        raise NoCMsFound("No CMs found could be connected to. Steam is likely down")

    async def fetch_refresh_token(self) -> str:
        client = self._state.client
        assert client.username
        assert client.password
        rsa_msg: auth.GetPasswordRsaPublicKeyResponse = await self.send_um_and_wait(
            auth.GetPasswordRsaPublicKeyRequest(client.username)
        )
        if rsa_msg.result != Result.OK:
            raise WSException(rsa_msg)

        begin_resp: auth.BeginAuthSessionViaCredentialsResponse = await self.send_um_and_wait(
            auth.BeginAuthSessionViaCredentialsRequest(
                account_name=client.username,
                encrypted_password=base64.b64encode(
                    rsa.RSAPublicNumbers(int(rsa_msg.publickey_exp, 16), int(rsa_msg.publickey_mod, 16))
                    .public_key()
                    .encrypt(client.password.encode(), padding.PKCS1v15())
                ).decode(),
                encryption_timestamp=rsa_msg.timestamp,
                platform_type=auth.EAuthTokenPlatformType.SteamClient,
                persistence=auth.ESessionPersistence.Persistent,
                website_id="Client",
            )
        )

        if not begin_resp.allowed_confirmations:
            raise AuthenticatorError("No valid auth session guard type was found")

        code_task = email_code_task = asyncio.get_running_loop().create_future()
        schedule_poll = False

        for allowed_confirmation in begin_resp.allowed_confirmations:
            match allowed_confirmation.confirmation_type:
                case auth.EAuthSessionGuardType.NONE:
                    poll_resp = await self.try_poll_auth_status(begin_resp)
                    break
                case auth.EAuthSessionGuardType.DeviceCode:
                    if not client.shared_secret:
                        print("Please enter a Steam guard code")
                    code_task = self.tg.create_task(self.update_auth_with_code(begin_resp, allowed_confirmation))
                case auth.EAuthSessionGuardType.EmailCode:
                    print("Please enter a confirmation code from your email")
                    email_code_task = self.tg.create_task(self.update_auth_with_code(begin_resp, allowed_confirmation))
                case auth.EAuthSessionGuardType.DeviceConfirmation:
                    schedule_poll = True
                    print("Confirm login this on your device")
                case auth.EAuthSessionGuardType.EmailConfirmation:
                    schedule_poll = True
                    print("Confirm login this via your email")
                case auth.EAuthSessionGuardType.MachineToken:
                    raise NotImplementedError("Machine tokens are not supported yet")
                case _:
                    raise NotImplementedError(
                        f"Unknown auth session guard type: {allowed_confirmation.confirmation_type}"
                    )
        else:
            poll_resp: auth.PollAuthSessionStatusResponse = await utils.race(
                code_task,
                email_code_task,
                self.poll_auth_status(begin_resp) if schedule_poll else asyncio.get_running_loop().create_future(),
            )
            assert poll_resp is not None

        self.client_id = poll_resp.new_client_id or begin_resp.client_id
        self._access_token = poll_resp.access_token
        return poll_resp.refresh_token

    async def update_auth_with_code(
        self,
        begin_resp: auth.BeginAuthSessionViaCredentialsResponse,
        allowed_confirmation: auth.AllowedConfirmation,
    ) -> auth.PollAuthSessionStatusResponse:
        type = allowed_confirmation.confirmation_type
        count = 0
        while True:
            code = await self._state.client.code()
            try:
                code_msg: auth.UpdateAuthSessionWithSteamGuardCodeResponse = await self.send_proto_and_wait(
                    auth.UpdateAuthSessionWithSteamGuardCodeRequest(
                        client_id=begin_resp.client_id,
                        steamid=begin_resp.steamid,
                        code=code,
                        code_type=type,
                    )
                )
            except ConnectionClosed:
                continue
            if code_msg.result == Result.TwoFactorCodeMismatch:
                count += 1
                if count > 3:
                    exc = WSException(code_msg)
                    if sys.version_info >= (3, 11):
                        msg = (
                            "Your clock is out of sync with the Steam servers or your shared secret is incorrect"
                            if self._state.client.shared_secret is not None
                            else "Your clock is likely out of sync with the Steam servers"
                        )
                        exc.add_note(msg)
                    raise exc
                if self._state.client.shared_secret is not None:
                    await asyncio.sleep(2)
                continue
            elif code_msg.result != Result.OK:
                raise WSException(code_msg)
            return await self.poll_auth_status(begin_resp)

    async def poll_auth_status(
        self, begin_resp: auth.BeginAuthSessionViaCredentialsResponse
    ) -> auth.PollAuthSessionStatusResponse:
        while True:
            resp = await self.try_poll_auth_status(begin_resp)
            if resp.refresh_token:
                return resp
            await asyncio.sleep(begin_resp.interval)

    async def try_poll_auth_status(
        self, begin_resp: auth.BeginAuthSessionViaCredentialsResponse
    ) -> auth.PollAuthSessionStatusResponse:
        poll_resp: auth.PollAuthSessionStatusResponse = await self.send_um_and_wait(
            auth.PollAuthSessionStatusRequest(
                client_id=begin_resp.client_id,
                request_id=begin_resp.request_id,
            )
        )
        if poll_resp.result != Result.OK:
            raise WSException(poll_resp)
        return poll_resp

    @utils.call_once(wait=True)
    async def access_token(self) -> str:
        try:
            return self._access_token
        except AttributeError:
            msg: auth.GenerateAccessTokenForAppResponse = await self.send_um_and_wait(
                auth.GenerateAccessTokenForAppRequest(self.refresh_token, self.id64)
            )
            if msg.result != Result.OK:
                raise WSException(msg) from None
            self._access_token = msg.access_token
            return msg.access_token

    @classmethod
    async def anonymous_login_from_client(
        cls, client: Client, cm_list: AsyncGenerator[CMServer, None] | None = None
    ) -> SteamWebSocket:
        state = client._state
        cm_list = cm_list or fetch_cm_list(state)
        async for cm in cm_list:
            log.info("Attempting to create a websocket connection to %s (load: %f)", cm.url, cm.weighted_load)
            socket = await cm.connect()
            log.debug("Connected to %s", cm.url)

            self = cls(state, socket, cm_list, cm)
            client.ws = self
            self._dispatch("connect")

            self.id64 = parse_id64(0, type=Type.AnonUser, universe=Universe.Public)

            async with self.poll():
                msg: login.CMsgClientLogonResponse = await self.send_proto_and_wait(
                    login.CMsgClientLogon(
                        protocol_version=PROTOCOL_VERSION,
                        client_package_version=1561159470,
                        client_language=state.language.api_name,
                    ),
                    check=lambda msg: isinstance(msg, login.CMsgClientLogonResponse),
                )
                if msg.result != Result.OK:
                    log.debug("Failed to login with result: %r", msg.result)
                    await self._state.handle_close()

                self.session_id = msg.header.session_id
                self.id64 = msg.header.steam_id
                self._state.cell_id = msg.cell_id

                self._keep_alive = KeepAliveHandler(
                    ws=self, interval=msg.heartbeat_seconds, loop=asyncio.get_running_loop()
                )
                self._keep_alive.start()
                log.debug("Heartbeat started.")

                state.login_complete.set()
                state.http.user = AnonymousClientUser(state, self.id64)  # type: ignore

                await self.send_proto(
                    login.CMsgClientServerTimestampRequest(client_request_timestamp=int(time.time() * 1000))
                )

                self._dispatch("login")
                log.debug("Logon completed")

            await client._handle_ready()
            return self
        raise NoCMsFound("No CMs found could be connected to. Steam is likely down")

    async def poll_event(self) -> None:
        try:
            message = await self.socket.receive()
            if message.type is aiohttp.WSMsgType.BINARY:
                return self.receive(message.data)
            if message.type is aiohttp.WSMsgType.ERROR:
                log.debug("Received %r", message)
                raise message.data
            if message.type in (aiohttp.WSMsgType.CLOSED, aiohttp.WSMsgType.CLOSE, aiohttp.WSMsgType.CLOSING):
                log.debug("Received %r", message)
                raise WebSocketClosure
            log.debug("Dropped unexpected message type: %r", message)
        except WebSocketClosure:
            await self._state.handle_close()

    def receive(self, message: bytes, /) -> None:
        emsg_value = READ_U32(message)
        try:
            msg = (
                ProtobufMessage().parse(message[4:], CLEAR_PROTO_BIT(emsg_value))
                if IS_PROTO(emsg_value)
                else Message().parse(message[4:], emsg_value)
            )
        except Exception as exc:
            return log.error(
                "Failed to deserialize message: %r, %r", EMsg(CLEAR_PROTO_BIT(emsg_value)), message, exc_info=exc
            )

        log.debug("Socket has received %r from the websocket.", msg)

        if hasattr(self, "_keep_alive"):
            self._keep_alive.tick()

        try:
            event_parser = self._state.parsers[msg.MSG]
        except (KeyError, TypeError):
            log.debug("Ignoring %r, no event handler", msg)
        else:
            try:
                result = event_parser(self._state, msg)
            except Exception:
                return traceback.print_exc()

            if isinstance(result, CoroutineType):
                self.tg.create_task(result, name=f"steam.py: {event_parser.__name__}")
        # remove the dispatched listener
        removed: list[int] = []
        for idx, entry in enumerate(self.listeners):
            if entry.msg != msg.MSG and entry.msg is not None:
                continue

            future = entry.future
            if future.cancelled():
                removed.append(idx)
                continue

            try:
                valid = entry.check(msg)
            except Exception as exc:
                future.set_exception(exc)
                removed.append(idx)
            else:
                if valid:
                    future.set_result(msg)
                    removed.append(idx)

        for idx in reversed(removed):
            del self.listeners[idx]

    async def send(self, data: bytes, /) -> None:
        try:
            await self.socket.send_bytes(data)
        except ConnectionResetError:
            await self._state.handle_close()

    async def send_proto(self, message: ProtoMsgs, /) -> None:
        message.header.steam_id = self.id64
        message.header.session_id = self.session_id

        await self.send(bytes(message))

    async def send_gc_message(self, msg: GCMsgs, /) -> int:  # for ext's to send GC messages
        app_id = msg.APP_ID
        message = client_server_2.CMsgGcClientToGC(
            appid=app_id,
            msgtype=SET_PROTO_BIT(msg.MSG) if isinstance(msg, GCProtobufMessage) else msg.MSG,
            payload=bytes(msg),
        )
        message.header.routing_app_id = app_id
        message.header.steam_id = self.id64
        message.header.job_id_source = self.next_gc_job_id

        log.debug("Sending GC message %r", msg)
        await self.send_proto(message)
        return message.header.job_id_source

    async def close(self, code: int = 1000) -> None:
        message = login.CMsgClientLogOff()
        message.header.steam_id = self.id64
        message.header.session_id = self.session_id
        await self.socket.close(code=code, message=bytes(message))

    @property
    def next_job_id(self) -> int:
        self._current_job_id = (self._current_job_id + 1) % 10000 or 1
        return self._current_job_id

    @property
    def next_gc_job_id(self) -> int:
        self._gc_current_job_id = (self._gc_current_job_id + 1) % 10000 or 1
        return self._gc_current_job_id

    async def send_um(self, um: UnifiedMessage, /) -> int:
        um.header.job_id_source = job_id = self.next_job_id
        await self.send_proto(um)
        return job_id

    async def send_um_and_wait(
        self,
        um: UnifiedMessage,
        /,
        check: Callable[[UnifiedMsgT], bool] | None = None,
    ) -> UnifiedMsgT:
        um.header.job_id_source = job_id = self.next_job_id
        check = check if check is not None else (lambda um: um.header.job_id_target == job_id)
        future = self.wait_for(emsg=EMsg.ServiceMethodSendToClient, check=check)
        await self.send_proto(um)
        return await future

    async def send_proto_and_wait(
        self, msg: ProtoMsgs, /, check: Callable[[ProtoMsgsT], bool] | None = None
    ) -> ProtoMsgsT:
        msg.header.job_id_source = job_id = self.next_job_id
        future = self.wait_for(
            emsg=None, check=check if check is not None else (lambda msg: msg.header.job_id_target == job_id)
        )
        await self.send_proto(msg)
        return await future

    async def send_gc_message_and_wait(
        self, msg: GCMsgs, /, check: Callable[[GCMsgProtoT], bool] | None = None
    ) -> GCMsgProtoT:
        msg.header.job_id_source = job_id = self.next_gc_job_id
        future = self.gc_wait_for(
            emsg=None,
            app_id=msg.APP_ID,
            check=check if check is not None else (lambda msg: msg.header.job_id_target == job_id),
        )
        await self.send_gc_message(msg)
        return await future

    async def change_presence(
        self,
        *,
        apps: list[client_server.CMsgClientGamesPlayedGamePlayed] | None = None,
        state: PersonaState | None = None,
        flags: PersonaStateFlag | None = None,
        ui_mode: UIMode | None = None,
        force_kick: bool = False,
    ) -> None:
        self._state._apps = apps or self._state._apps
        self._state._state = state or self._state._state
        self._state._ui_mode = ui_mode or self._state._ui_mode
        self._state._flags = flags or self._state._flags
        self._state._force_kick = force_kick

        if force_kick:
            log.debug("Kicking any currently playing sessions")
            await self.send_proto(client_server_2.CMsgClientKickPlayingSession())
        if apps is not None:
            apps_msg = client_server.CMsgClientGamesPlayed(games_played=apps)
            log.debug("Sending %r to change activity", apps_msg)
            await self.send_proto(apps_msg)
        if state is not None or flags is not None:
            state_msg = friends.CMsgClientChangeStatus(
                persona_state=state or self._state._state, persona_state_flags=flags or self._state._flags
            )
            log.debug("Sending %r to change state", state_msg)
            await self.send_proto(state_msg)
        if ui_mode is not None:
            ui_mode_msg = client_server_2.CMsgClientUiMode(uimode=ui_mode)
            log.debug("Sending %r to change UI mode", ui_mode_msg)
            await self.send_proto(ui_mode_msg)

    async def fetch_users(
        self, user_id64s: Iterable[ID64]
    ) -> AsyncGenerator[friends.CMsgClientPersonaStateFriend, None]:
        futs: list[asyncio.Future[friends.CMsgClientPersonaState]] = []
        user_id64s = dict.fromkeys(user_id64s)

        def callback(msg: friends.CMsgClientPersonaState, user_id64: ID64) -> bool:
            return any(friend.friendid == user_id64 for friend in msg.friends)

        for user_id64_chunk in utils.as_chunks(user_id64s, 100):
            futs += [
                self.wait_for(friends.CMsgClientPersonaState, check=partial(callback, user_id64=user_id64))
                for user_id64 in user_id64_chunk
            ]
            await self.send_proto(
                friends.CMsgClientRequestFriendData(
                    # enum EClientPersonaStateFlag {
                    #     Status = 1;
                    #     PlayerName = 2;
                    #     QueryPort = 4;
                    #     SourceID = 8;
                    #     Presence = 16;
                    #     LastSeen = 64;
                    #     UserClanRank = 128;
                    #     ExtraInfo = 256;
                    #     DataBlob = 512;
                    #     ClanData = 1024;
                    #     Facebook = 2048;
                    #     RichPresence = 4096;
                    #     Broadcast = 8192;
                    #     Watching = 16384;
                    # };
                    persona_state_requested=0b1111111011111,  # all except watching and broadcast (for now)
                    friends=list(user_id64_chunk),
                ),
            )

        for msg in await asyncio.wait_for(asyncio.gather(*futs), timeout=60):
            if msg.result not in (Result.OK, Result.Invalid):  # not sure if checking this is even useful
                raise WSException(msg)

            for user in msg.friends:
                if user.friendid in user_id64s:
                    yield user



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Generic

from .abc import Awardable, Commentable, _CommentableKwargs
from .types.user import UserT
from .utils import DateTime

if TYPE_CHECKING:
    from datetime import datetime

    from ._const import ReadOnly
    from .app import PartialApp
    from .state import ConnectionState
    from .types.id import PostID


@dataclass(repr=False, slots=True, unsafe_hash=True)
class Post(Awardable["PostID"], Commentable, Generic[UserT]):
    """Represents a post on Steam Community."""

    _state: ConnectionState = field(compare=False, hash=False)
    id: ReadOnly[PostID]
    """The ID of this post."""
    content: str = field(compare=False, hash=False)
    """The content of this post."""
    author: UserT
    """The author of this post."""
    app: PartialApp | None = field(compare=False, hash=False)
    """The app this post is for, if any."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} id={self.id} author={self.author!r} app={self.app!r}>"

    async def delete(self) -> None:
        """Delete this post."""
        await self._state.delete_user_post(self.id)

    async def upvote(self) -> None:
        """Upvote this post."""
        await self._state.http.vote_on_user_post(self.author.id64, self.id, 1)

    async def downvote(self) -> None:
        """Un-upvote this post (this is different from other downvoting systems)."""
        await self._state.http.vote_on_user_post(self.author.id64, self.id, 0)

    @property
    def created_at(self) -> datetime:
        """The time this post was created."""
        return DateTime.from_timestamp(self.id)  # yes, really

    @property
    def url(self) -> str:
        """The URL of this post."""
        return f"{self.author.community_url}/status/{self.id}"

    @property
    def _commentable_kwargs(self) -> _CommentableKwargs:
        return {
            "id64": self.author.id64,
            "forum_id": self.id,
        }



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from collections.abc import AsyncGenerator, Mapping
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TYPE_CHECKING, Final, Protocol, cast, overload

from yarl import URL

from .app import PartialApp
from .models import _IOMixin, _IOMixinNoOpen
from .protobufs.chat import EChatRoomMessageReactionType
from .utils import DateTime

if TYPE_CHECKING:
    from datetime import datetime

    from aiohttp import StreamReader

    from .message import Message
    from .protobufs.friend_messages import EMessageReactionType
    from .protobufs.friends import (
        CMsgClientEmoticonListEffect as ClientEffectProto,
        CMsgClientEmoticonListEmoticon as ClientEmoticonProto,
        CMsgClientEmoticonListSticker as ClientStickerProto,
    )
    from .state import ConnectionState
    from .types.user import Author


__all__ = (
    "Award",
    "AwardReaction",
    "PartialMessageReaction",
    "MessageReaction",
    "Emoticon",
    "Sticker",
    "ClientEmoticon",
    "ClientSticker",
    "ClientEffect",
)


class ReactionProtocol(Protocol):
    reactionid: int
    count: int


@dataclass(slots=True)
class _Reaction:
    reactionid: int
    count: int


BASE_REACTION_URL: Final = "https://store.cloudflare.steamstatic.com/public/images/loyalty/reactions/{type}/{id}.png"
BASE_ECONOMY_URL: Final = URL("https://community.akamai.steamstatic.com/economy")
AWARD_ID_TO_NAME: Final = cast(Mapping[int, str], {
    1: "Deep Thoughts",
    2: "Heartwarming",
    3: "Hilarious",
    4: "Hot Take",
    5: "Poetry",
    6: "Extra Helpful",
    7: "Gotta Have It",
    8: "Michelangelo",
    9: "Treasure",
    10: "Mind Blown",
    11: "Golden Unicorn",
    12: "Mad Scientist",
    13: "Clever",
    14: "Warm Blanket",
    15: "Saucy",
    16: "Slow Clap",
    17: "Take My Points",
    18: "Wholesome",
    19: "Jester",
    20: "Fancy Pants",
    21: "Whoa",
    22: "Super Star",
    23: "Wild",
})  # fmt: skip


class Award(_IOMixinNoOpen):
    """Represents an award."""

    def __init__(self, state: ConnectionState, id: int):
        self._state = state
        self.id = id
        """The ID of the award."""
        self.name = AWARD_ID_TO_NAME.get(self.id, f"Unknown Award {self.id}")
        """The english localised name of the award."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} id={self.id}>"

    @property
    def url(self) -> str:
        """The reaction's icon URL."""
        return BASE_REACTION_URL.format(type="still", id=self.id)

    @property
    def animated_url(self) -> str:
        """The reaction's icon animated URL."""
        return BASE_REACTION_URL.format(type="animated", id=self.id)

    @asynccontextmanager
    async def open(self, *, animated: bool = False) -> AsyncGenerator[StreamReader, None]:
        async with self._state.http._session.get(self.animated_url if animated else self.url) as r:
            yield r.content


class AwardReaction:
    """Represents an award on user generated content."""

    def __init__(self, state: ConnectionState, proto: ReactionProtocol):
        self._state = state
        self.award = Award(state, proto.reactionid)
        """The award given to the user generated content."""
        self.count = proto.count
        """The reactions count on the post."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} award={self.award!r} count={self.count}>"


@dataclass(slots=True)
class PartialMessageReaction:
    """Represents a partial reaction to a message."""

    _state: ConnectionState
    message: Message
    """The message the reaction applies to. """
    emoticon: Emoticon | None
    """The emoticon that was reacted with."""
    sticker: Sticker | None
    """The sticker that was reacted with."""

    if TYPE_CHECKING:

        @overload
        def __init__(self, _state: ConnectionState, message: Message, emoticon: Emoticon, sticker: None) -> None:  # type: ignore
            ...

        @overload
        def __init__(self, _state: ConnectionState, message: Message, emoticon: None, sticker: Sticker) -> None: ...

    def __repr__(self) -> str:
        return (
            f"<{self.__class__.__name__} message={self.message!r} emoticon={self.emoticon!r} sticker={self.sticker!r}>"
        )


@dataclass(slots=True, eq=False, repr=False)
class MessageReaction(PartialMessageReaction):
    """Represents a reaction to a message."""

    user: Author
    """The user that reacted to the message."""
    created_at: datetime | None = None
    """The time the reaction was added to the message."""
    ordinal: int | None = None
    """The ordinal of the the message."""

    if TYPE_CHECKING:

        @overload
        def __init__(  # type: ignore
            self,
            _state: ConnectionState,
            message: Message,
            emoticon: Emoticon,
            sticker: None,
            user: Author,
            created_at: datetime | None = None,
            ordinal: int | None = None,
        ) -> None: ...

        @overload
        def __init__(
            self,
            _state: ConnectionState,
            message: Message,
            emoticon: None,
            sticker: Sticker,
            user: Author,
            created_at: datetime | None = None,
            ordinal: int | None = None,
        ) -> None: ...

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False

        return (
            self.message == other.message
            and self.emoticon == other.emoticon
            and self.sticker == other.sticker
            and self.user == other.user
        )

    def __hash__(self) -> int:
        return hash((self.message, self.emoticon, self.sticker, self.user))

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} message={self.message!r} emoticon={self.emoticon!r} sticker={self.sticker!r} user={self.user!r}>"


class BaseEmoticon(_IOMixin):
    __slots__ = ("_state", "name")

    def __init__(self, state: ConnectionState, name: str):
        self._state = state
        self.name = name.removeprefix(":").removesuffix(":")  # :emoji_name:
        """The name of this emoticon."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r}>"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseEmoticon):
            return False
        if (isinstance(self, Emoticon) and isinstance(other, Emoticon)) or (
            isinstance(self, Sticker) and isinstance(other, Sticker)
        ):
            return self.name == other.name

        return False

    def __hash__(self) -> int:
        return hash(self.name)

    @property
    def url(self) -> str:
        raise NotImplementedError


EMOTICON_TYPE = EChatRoomMessageReactionType.Emoticon
STICKER_TYPE = EChatRoomMessageReactionType.Sticker

if TYPE_CHECKING:  # until actual intersection types are added this will have to do
    assert isinstance(EMOTICON_TYPE, EMessageReactionType)
    assert isinstance(STICKER_TYPE, EMessageReactionType)


class Emoticon(BaseEmoticon):
    """Represents an emoticon in chat.

    .. container:: operations

        .. describe:: x == y

            Checks if two emoticons are equal.

        .. describe:: str(x)

            The string to send this emoticon in chat.

        .. describe:: hash(x)

            Returns the emoticon's hash.
    """

    __slots__ = ()
    _TYPE: Final = EMOTICON_TYPE

    def __str__(self) -> str:
        return f":{self.name}:"

    @property
    def url(self) -> str:
        """The URL for this emoticon."""
        return str(BASE_ECONOMY_URL / "emoticonlarge" / self.name)

    async def app(self) -> PartialApp[str]:
        """Fetches this emoticon's associated app."""
        data = await self._state.http.get(BASE_ECONOMY_URL / "emoticonhoverjson" / self.name)
        return PartialApp(self._state, id=data["appid"], name=data["app_name"])


class Sticker(BaseEmoticon):
    """Represents a sticker in chat.

    .. container:: operations

        .. describe:: x == y

            Checks if two stickers are equal.

        .. describe:: str(x)

            The way to send this sticker in chat.

        .. describe:: hash(x)

            Returns the stickers's hash.

    Note
    ----
    Unlike :class:`Emoticon` this can only be sent in a message by itself.
    """

    __slots__ = ()
    _TYPE: Final = STICKER_TYPE

    def __str__(self) -> str:
        return f"/sticker {self.name}"

    @property
    def url(self) -> str:
        """The URL for this sticker."""
        return str(BASE_ECONOMY_URL / "sticker" / self.name)

    async def app(self) -> PartialApp[None]:
        """Fetches this sticker's associated app."""
        data = await self._state.http.get(BASE_ECONOMY_URL / "stickerjson" / self.name)
        return PartialApp(self._state, id=data["appid"])


class BaseClientEmoticon(BaseEmoticon):
    __slots__ = ("count", "use_count", "last_used", "received_at")

    def __init__(self, state: ConnectionState, proto: ClientEmoticonProto | ClientStickerProto):
        super().__init__(state, proto.name)
        self.count = proto.count
        """The number of times this emoticon can be used."""
        self.use_count = proto.use_count
        """The number of times this emoticon has been used."""
        self.last_used = DateTime.from_timestamp(proto.time_last_used)
        """The last time this emoticon was used."""
        self.received_at = DateTime.from_timestamp(proto.time_received)
        """The time this emoticon was received."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} count={self.count!r} use_count={self.use_count!r}>"


class ClientEmoticon(BaseClientEmoticon, Emoticon):
    """Represents an :class:`Emoticon` the :class:`ClientUser` owns."""

    __slots__ = ()


class ClientSticker(BaseClientEmoticon, Sticker):
    """Represents a :class:`Sticker` the :class:`ClientUser` owns."""

    __slots__ = ("_app_id",)

    def __init__(self, state: ConnectionState, proto: ClientStickerProto):
        super().__init__(state, proto)
        self._app_id = proto.appid

    async def app(self) -> PartialApp[None]:
        return PartialApp(self._state, id=self._app_id)  # no point fetching cause endpoint doesn't return name


class Effect(BaseEmoticon): ...


class ClientEffect(Effect):
    __slots__ = (
        "count",
        "received_at",
        "infinite_use",
        "_app_id",
    )

    def __init__(self, state: ConnectionState, proto: ClientEffectProto):
        super().__init__(state, proto.name)
        self.count = proto.count
        """The number of times this emoticon can be used."""
        self.received_at = DateTime.from_timestamp(proto.time_received)
        """The time this emoticon was received."""
        self.infinite_use = proto.infinite_use
        self._app_id = proto.appid

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} name={self.name!r} count={self.count!r}>"

    def __str__(self) -> str:
        return f"/roomeffect {self.name}"

    async def app(self) -> PartialApp[None]:
        """Fetches this effect's associated app."""
        return PartialApp(self._state, id=self._app_id)  # no point fetching cause endpoint doesn't return name


# https://cdn.akamai.steamstatic.com/steamcommunity/public/assets/winter2019/roomeffects/96px/firework.png
# Message.sticker?/roomeffect



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_chat.proto
# plugin: python-betterproto
# Last updated 09/09/2021

from dataclasses import dataclass

import betterproto

from .friends import CMsgClientPersonaStateFriend
from .msg import UnifiedMessage


class EChatRoomJoinState(betterproto.Enum):
    Default = 0
    NONE = 1
    Joined = 2
    TestInvalid = 99


class EChatRoomGroupRank(betterproto.Enum):
    Default = 0
    Viewer = 10
    Guest = 15
    Member = 20
    Moderator = 30
    Officer = 40
    Owner = 50
    TestInvalid = 99


class EChatRoomNotificationLevel(betterproto.Enum):
    Invalid = 0
    NONE = 1
    MentionMe = 2
    MentionAll = 3
    AllMessages = 4


class EChatRoomServerMessage(betterproto.Enum):
    Invalid = 0
    RenameChatRoom = 1
    Joined = 2
    Parted = 3
    Kicked = 4
    Invited = 5
    InviteDismissed = 8
    ChatRoomTaglineChanged = 9
    ChatRoomAvatarChanged = 10
    AppCustom = 11


class EChatRoomMessageReactionType(betterproto.Enum):
    Invalid = 0
    Emoticon = 1
    Sticker = 2


class EChatRoomMemberStateChange(betterproto.Enum):
    Invalid = 0
    Joined = 1
    Parted = 2
    Kicked = 3
    Invited = 4
    RankChanged = 7
    InviteDismissed = 8
    Muted = 9
    Banned = 10
    RolesChanged = 12


@dataclass(eq=False, repr=False)
class RequestFriendPersonaStatesRequest(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RequestFriendPersonaStatesResponse(betterproto.Message):
    pass


class CreateChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.CreateChatRoomGroup"):
    steamid_partner: int = betterproto.fixed64_field(1)
    steamid_invited: int = betterproto.fixed64_field(2)
    name: str = betterproto.string_field(3)
    steamid_invitees: list[int] = betterproto.fixed64_field(4)
    watching_broadcast_accountid: int = betterproto.uint32_field(6)
    watching_broadcast_channel_id: int = betterproto.uint64_field(7)


@dataclass(eq=False, repr=False)
class Role(betterproto.Message):
    role_id: int = betterproto.uint64_field(1)
    name: str = betterproto.string_field(2)
    ordinal: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class RoleActions(betterproto.Message):
    role_id: int = betterproto.uint64_field(1)
    can_create_rename_delete_channel: bool = betterproto.bool_field(2)
    can_kick: bool = betterproto.bool_field(3)
    can_ban: bool = betterproto.bool_field(4)
    can_invite: bool = betterproto.bool_field(5)
    can_change_tagline_avatar_name: bool = betterproto.bool_field(6)
    can_chat: bool = betterproto.bool_field(7)
    can_view_history: bool = betterproto.bool_field(8)
    can_change_group_roles: bool = betterproto.bool_field(9)
    can_change_user_roles: bool = betterproto.bool_field(10)
    can_mention_all: bool = betterproto.bool_field(11)
    can_set_watching_broadcast: bool = betterproto.bool_field(12)


@dataclass(eq=False, repr=False)
class PartyBeacon(betterproto.Message):
    app_id: int = betterproto.uint32_field(1)
    steamid_owner: int = betterproto.fixed64_field(2)
    beacon_id: int = betterproto.fixed64_field(3)
    game_metadata: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class GroupHeaderState(betterproto.Message):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_name: str = betterproto.string_field(2)
    clanid: int = betterproto.uint32_field(13)
    accountid_owner: int = betterproto.uint32_field(14)
    appid: int = betterproto.uint32_field(21)
    tagline: str = betterproto.string_field(15)
    avatar_sha: bytes = betterproto.bytes_field(16)
    default_role_id: int = betterproto.uint64_field(17)
    roles: "list[Role]" = betterproto.message_field(18)
    role_actions: "list[RoleActions]" = betterproto.message_field(19)
    watching_broadcast_accountid: int = betterproto.uint32_field(20)
    party_beacons: "list[PartyBeacon]" = betterproto.message_field(22)
    watching_broadcast_channel_id: int = betterproto.uint64_field(23)
    active_minigame_id: int = betterproto.uint64_field(24)
    avatar_ugc_url: str = betterproto.string_field(25)
    disabled: bool = betterproto.bool_field(26)


@dataclass(eq=False, repr=False)
class Member(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    state: "EChatRoomJoinState" = betterproto.enum_field(3)
    rank: "EChatRoomGroupRank" = betterproto.enum_field(4)
    time_kick_expire: int = betterproto.uint32_field(6)
    role_ids: list[int] = betterproto.uint64_field(7)


@dataclass(eq=False, repr=False)
class State(betterproto.Message):
    chat_id: int = betterproto.uint64_field(1)
    chat_name: str = betterproto.string_field(2)
    voice_allowed: bool = betterproto.bool_field(3)
    members_in_voice: list[int] = betterproto.uint32_field(4)
    time_last_message: int = betterproto.uint32_field(5)
    sort_order: int = betterproto.uint32_field(6)
    last_message: str = betterproto.string_field(7)
    accountid_last_message: int = betterproto.uint32_field(8)


@dataclass(eq=False, repr=False)
class GroupState(betterproto.Message):
    header_state: "GroupHeaderState" = betterproto.message_field(1)
    members: "list[Member]" = betterproto.message_field(2)
    default_chat_id: int = betterproto.uint64_field(4)
    chat_rooms: "list[State]" = betterproto.message_field(5)
    kicked: "list[Member]" = betterproto.message_field(7)


@dataclass(eq=False, repr=False)
class ChatRoomState(betterproto.Message):
    chat_id: int = betterproto.uint64_field(1)
    time_joined: int = betterproto.uint32_field(2)
    time_last_ack: int = betterproto.uint32_field(3)
    desktop_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(4)
    mobile_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(5)
    time_last_mention: int = betterproto.uint32_field(6)
    unread_indicator_muted: bool = betterproto.bool_field(7)
    time_first_unread: int = betterproto.uint32_field(8)


@dataclass(eq=False, repr=False)
class ChatRoomGroupState(betterproto.Message):
    chat_group_id: int = betterproto.uint64_field(1)
    time_joined: int = betterproto.uint32_field(2)
    user_chat_room_state: "list[ChatRoomState]" = betterproto.message_field(3)
    desktop_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(4)
    mobile_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(5)
    time_last_group_ack: int = betterproto.uint32_field(6)
    unread_indicator_muted: bool = betterproto.bool_field(7)


class CreateChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.CreateChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    state: "GroupState" = betterproto.message_field(2)
    user_chat_state: "ChatRoomGroupState" = betterproto.message_field(3)


class SaveChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.SaveChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    name: str = betterproto.string_field(2)


class SaveChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.SaveChatRoomGroup"):
    pass


class RenameChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.RenameChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    name: str = betterproto.string_field(2)


class RenameChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.RenameChatRoomGroup"):
    name: str = betterproto.string_field(1)


class SetChatRoomGroupTaglineRequest(UnifiedMessage, um_name="ChatRoom.SetChatRoomGroupTagline"):
    chat_group_id: int = betterproto.uint64_field(1)
    tagline: str = betterproto.string_field(2)


class SetChatRoomGroupTaglineResponse(UnifiedMessage, um_name="ChatRoom.SetChatRoomGroupTagline"):
    pass


class SetChatRoomGroupAvatarRequest(UnifiedMessage, um_name="ChatRoom.SetChatRoomGroupAvatar"):
    chat_group_id: int = betterproto.uint64_field(1)
    avatar_sha: bytes = betterproto.bytes_field(2)


class SetChatRoomGroupAvatarResponse(UnifiedMessage, um_name="ChatRoom.SetChatRoomGroupAvatar"):
    pass


class SetChatRoomGroupWatchingBroadcastRequest(UnifiedMessage, um_name="ChatRoom.SetChatRoomGroupWatchingBroadcast"):
    chat_group_id: int = betterproto.uint64_field(1)
    watching_broadcast_accountid: int = betterproto.uint32_field(2)
    watching_broadcast_channel_id: int = betterproto.uint64_field(3)


class SetChatRoomGroupWatchingBroadcastResponse(UnifiedMessage, um_name="ChatRoom.SetChatRoomGroupWatchingBroadcast"):
    pass


class JoinMiniGameForChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.JoinMiniGameForChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)


class JoinMiniGameForChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.JoinMiniGameForChatRoomGroup"):
    minigame_id: int = betterproto.uint64_field(1)


class EndMiniGameForChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.EndMiniGameForChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    minigame_id: int = betterproto.uint64_field(3)


class EndMiniGameForChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.EndMiniGameForChatRoomGroup"):
    pass


class MuteUserRequest(UnifiedMessage, um_name="ChatRoom.MuteUserInGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    steamid: int = betterproto.fixed64_field(2)
    expiration: int = betterproto.int32_field(3)


class MuteUserResponse(UnifiedMessage, um_name="ChatRoom.MuteUserInGroup"):
    pass


class KickUserRequest(UnifiedMessage, um_name="ChatRoom.KickUserFromGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    steamid: int = betterproto.fixed64_field(2)
    expiration: int = betterproto.int32_field(3)


class KickUserResponse(UnifiedMessage, um_name="ChatRoom.KickUserFromGroup"):
    pass


class SetUserBanStateRequest(UnifiedMessage, um_name="ChatRoom.SetUserBanState"):
    chat_group_id: int = betterproto.uint64_field(1)
    steamid: int = betterproto.fixed64_field(2)
    ban_state: bool = betterproto.bool_field(3)


class SetUserBanStateResponse(UnifiedMessage, um_name="ChatRoom.SetUserBanState"):
    pass


class RevokeInviteRequest(UnifiedMessage, um_name="ChatRoom.RevokeInviteToGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    steamid: int = betterproto.fixed64_field(2)


class RevokeInviteResponse(UnifiedMessage, um_name="ChatRoom.RevokeInviteToGroup"):
    pass


class CreateRoleRequest(UnifiedMessage, um_name="ChatRoom.CreateRole"):
    chat_group_id: int = betterproto.uint64_field(1)
    name: str = betterproto.string_field(2)


class CreateRoleResponse(UnifiedMessage, um_name="ChatRoom.CreateRole"):
    actions: "RoleActions" = betterproto.message_field(2)


class GetRolesRequest(UnifiedMessage, um_name="ChatRoom.GetRoles"):
    chat_group_id: int = betterproto.uint64_field(1)


class GetRolesResponse(UnifiedMessage, um_name="ChatRoom.GetRoles"):
    roles: "list[Role]" = betterproto.message_field(1)


class RenameRoleRequest(UnifiedMessage, um_name="ChatRoom.RenameRole"):
    chat_group_id: int = betterproto.uint64_field(1)
    role_id: int = betterproto.uint64_field(2)
    name: str = betterproto.string_field(3)


class RenameRoleResponse(UnifiedMessage, um_name="ChatRoom.RenameRole"):
    pass


class ReorderRoleRequest(UnifiedMessage, um_name="ChatRoom.ReorderRole"):
    chat_group_id: int = betterproto.uint64_field(1)
    role_id: int = betterproto.uint64_field(2)
    ordinal: int = betterproto.uint32_field(3)


class ReorderRoleResponse(UnifiedMessage, um_name="ChatRoom.ReorderRole"):
    pass


class DeleteRoleRequest(UnifiedMessage, um_name="ChatRoom.DeleteRole"):
    chat_group_id: int = betterproto.uint64_field(1)
    role_id: int = betterproto.uint64_field(2)


class DeleteRoleResponse(UnifiedMessage, um_name="ChatRoom.DeleteRole"):
    pass


class GetRoleActionsRequest(UnifiedMessage, um_name="ChatRoom.GetRoleActions"):
    chat_group_id: int = betterproto.uint64_field(1)
    role_id: int = betterproto.uint64_field(2)


class GetRoleActionsResponse(UnifiedMessage, um_name="ChatRoom.GetRoleActions"):
    actions: "list[RoleActions]" = betterproto.message_field(1)


class ReplaceRoleActionsRequest(UnifiedMessage, um_name="ChatRoom.ReplaceRoleActions"):
    chat_group_id: int = betterproto.uint64_field(1)
    role_id: int = betterproto.uint64_field(2)
    actions: "RoleActions" = betterproto.message_field(4)


class ReplaceRoleActionsResponse(UnifiedMessage, um_name="ChatRoom.ReplaceRoleActions"):
    pass


class AddRoleToUserRequest(UnifiedMessage, um_name="ChatRoom.AddRoleToUser"):
    chat_group_id: int = betterproto.uint64_field(1)
    role_id: int = betterproto.uint64_field(3)
    steamid: int = betterproto.fixed64_field(4)


class AddRoleToUserResponse(UnifiedMessage, um_name="ChatRoom.AddRoleToUser"):
    pass


class GetRolesForUserRequest(UnifiedMessage, um_name="ChatRoom.GetRolesForUser"):
    chat_group_id: int = betterproto.uint64_field(1)
    steamid: int = betterproto.fixed64_field(3)


class GetRolesForUserResponse(UnifiedMessage, um_name="ChatRoom.GetRolesForUser"):
    role_ids: list[int] = betterproto.uint64_field(1)


class DeleteRoleFromUserRequest(UnifiedMessage, um_name="ChatRoom.DeleteRoleFromUser"):
    chat_group_id: int = betterproto.uint64_field(1)
    role_id: int = betterproto.uint64_field(3)
    steamid: int = betterproto.fixed64_field(4)


class DeleteRoleFromUserResponse(UnifiedMessage, um_name="ChatRoom.DeleteRoleFromUser"):
    pass


class JoinChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.JoinChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    invite_code: str = betterproto.string_field(2)
    chat_id: int = betterproto.uint64_field(3)


class JoinChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.JoinChatRoomGroup"):
    state: "GroupState" = betterproto.message_field(1)
    user_chat_state: "ChatRoomGroupState" = betterproto.message_field(3)
    join_chat_id: int = betterproto.uint64_field(4)
    time_expire: int = betterproto.uint32_field(5)


class InviteFriendToChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.InviteFriendToChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)
    steamid: int = betterproto.fixed64_field(2)
    chat_id: int = betterproto.uint64_field(3)
    skip_friendsui_check: bool = betterproto.bool_field(4)


class InviteFriendToChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.InviteFriendToChatRoomGroup"):
    pass


class LeaveChatRoomGroupRequest(UnifiedMessage, um_name="ChatRoom.LeaveChatRoomGroup"):
    chat_group_id: int = betterproto.uint64_field(1)


class LeaveChatRoomGroupResponse(UnifiedMessage, um_name="ChatRoom.LeaveChatRoomGroup"):
    pass


class CreateChatRoomRequest(UnifiedMessage, um_name="ChatRoom.CreateChatRoom"):
    chat_group_id: int = betterproto.uint64_field(1)
    name: str = betterproto.string_field(2)
    allow_voice: bool = betterproto.bool_field(3)


class CreateChatRoomResponse(UnifiedMessage, um_name="ChatRoom.CreateChatRoom"):
    chat_room: "State" = betterproto.message_field(1)


class DeleteChatRoomRequest(UnifiedMessage, um_name="ChatRoom.DeleteChatRoom"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)


class DeleteChatRoomResponse(UnifiedMessage, um_name="ChatRoom.DeleteChatRoom"):
    pass


class RenameChatRoomRequest(UnifiedMessage, um_name="ChatRoom.RenameChatRoom"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    name: str = betterproto.string_field(3)


class RenameChatRoomResponse(UnifiedMessage, um_name="ChatRoom.RenameChatRoom"):
    pass


class ReorderChatRoomRequest(UnifiedMessage, um_name="ChatRoom.ReorderChatRoom"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    move_after_chat_id: int = betterproto.uint64_field(3)


class ReorderChatRoomResponse(UnifiedMessage, um_name="ChatRoom.ReorderChatRoom"):
    pass


class SendChatMessageRequest(UnifiedMessage, um_name="ChatRoom.SendChatMessage"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    message: str = betterproto.string_field(3)
    echo_to_sender: bool = betterproto.bool_field(4)


class SendChatMessageResponse(UnifiedMessage, um_name="ChatRoom.SendChatMessage"):
    modified_message: str = betterproto.string_field(1)
    server_timestamp: int = betterproto.uint32_field(2)
    ordinal: int = betterproto.uint32_field(3)
    message_without_bb_code: str = betterproto.string_field(4)


class JoinVoiceChatRequest(UnifiedMessage, um_name="ChatRoom.JoinVoiceChat"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)


class JoinVoiceChatResponse(UnifiedMessage, um_name="ChatRoom.JoinVoiceChat"):
    voice_chatid: int = betterproto.uint64_field(1)


class LeaveVoiceChatRequest(UnifiedMessage, um_name="ChatRoom.LeaveVoiceChat"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)


class LeaveVoiceChatResponse(UnifiedMessage, um_name="ChatRoom.LeaveVoiceChat"):
    pass


class GetMessageHistoryRequest(UnifiedMessage, um_name="ChatRoom.GetMessageHistory"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    last_time: int = betterproto.uint32_field(3)
    last_ordinal: int = betterproto.uint32_field(4)
    start_time: int = betterproto.uint32_field(5)
    start_ordinal: int = betterproto.uint32_field(6)
    max_count: int = betterproto.uint32_field(7)


@dataclass(eq=False, repr=False)
class ServerMessage(betterproto.Message):
    message: "EChatRoomServerMessage" = betterproto.enum_field(1)
    string_param: str = betterproto.string_field(2)
    accountid_param: int = betterproto.uint32_field(3)


class GetMessageHistoryResponse(UnifiedMessage, um_name="ChatRoom.GetMessageHistory"):
    messages: "list[GetMessageHistoryResponseChatMessage]" = betterproto.message_field(1)
    more_available: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class GetMessageHistoryResponseChatMessage(betterproto.Message):
    sender: int = betterproto.uint32_field(1)
    server_timestamp: int = betterproto.uint32_field(2)
    message: str = betterproto.string_field(3)
    ordinal: int = betterproto.uint32_field(4)
    server_message: "ServerMessage" = betterproto.message_field(5)
    deleted: bool = betterproto.bool_field(6)
    reactions: "list[GetMessageHistoryResponseChatMessageMessageReaction]" = betterproto.message_field(7)


@dataclass(eq=False, repr=False)
class GetMessageHistoryResponseChatMessageMessageReaction(betterproto.Message):
    reaction_type: "EChatRoomMessageReactionType" = betterproto.enum_field(1)
    reaction: str = betterproto.string_field(2)
    num_reactors: int = betterproto.uint32_field(3)
    has_user_reacted: bool = betterproto.bool_field(4)


class GetMyChatRoomGroupsRequest(UnifiedMessage, um_name="ChatRoom.GetMyChatRoomGroups"):
    pass


class GetChatRoomGroupSummaryResponse(UnifiedMessage, um_name="ChatRoom.GetChatRoomGroupSummary"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_group_name: str = betterproto.string_field(2)
    active_member_count: int = betterproto.uint32_field(3)
    active_voice_member_count: int = betterproto.uint32_field(4)
    default_chat_id: int = betterproto.uint64_field(5)
    chat_rooms: "list[State]" = betterproto.message_field(6)
    clanid: int = betterproto.uint32_field(7)
    chat_group_tagline: str = betterproto.string_field(8)
    accountid_owner: int = betterproto.uint32_field(9)
    top_members: list[int] = betterproto.uint32_field(10)
    chat_group_avatar_sha: bytes = betterproto.bytes_field(11)
    rank: "EChatRoomGroupRank" = betterproto.enum_field(12)
    default_role_id: int = betterproto.uint64_field(13)
    role_ids: list[int] = betterproto.uint64_field(14)
    role_actions: "list[RoleActions]" = betterproto.message_field(15)
    watching_broadcast_accountid: int = betterproto.uint32_field(16)
    appid: int = betterproto.uint32_field(17)
    party_beacons: "list[PartyBeacon]" = betterproto.message_field(18)
    watching_broadcast_channel_id: int = betterproto.uint64_field(19)
    active_minigame_id: int = betterproto.uint64_field(20)
    avatar_ugc_url: str = betterproto.string_field(21)
    disabled: bool = betterproto.bool_field(22)


@dataclass(eq=False, repr=False)
class SummaryPair(betterproto.Message):
    user_chat_group_state: "ChatRoomGroupState" = betterproto.message_field(1)
    group_summary: "GetChatRoomGroupSummaryResponse" = betterproto.message_field(2)


class GetMyChatRoomGroupsResponse(UnifiedMessage, um_name="ChatRoom.GetMyChatRoomGroups"):
    chat_room_groups: "list[SummaryPair]" = betterproto.message_field(1)


class GetChatRoomGroupStateRequest(UnifiedMessage, um_name="ChatRoom.GetChatRoomGroupState"):
    chat_group_id: int = betterproto.uint64_field(1)


class GetChatRoomGroupStateResponse(UnifiedMessage, um_name="ChatRoom.GetChatRoomGroupState"):
    state: "GroupState" = betterproto.message_field(1)


class GetChatRoomGroupSummaryRequest(UnifiedMessage, um_name="ChatRoom.GetChatRoomGroupSummary"):
    chat_group_id: int = betterproto.uint64_field(1)


class SetAppChatRoomGroupForceActiveRequest(UnifiedMessage, um_name="ChatRoom.SetAppChatRoomGroupForceActive"):
    chat_group_id: int = betterproto.uint64_field(1)
    requesting_app_id: int = betterproto.uint32_field(2)


class SetAppChatRoomGroupForceActiveResponse(UnifiedMessage, um_name="ChatRoom.SetAppChatRoomGroupForceActive"):
    eresult: int = betterproto.uint32_field(1)
    accounts_in_channel: list[int] = betterproto.uint32_field(2)


class SetAppChatRoomGroupStopForceActiveNotification(
    UnifiedMessage, um_name="ChatRoom.SetAppChatRoomGroupStopForceActive"
):
    chat_group_id: int = betterproto.uint64_field(1)
    requesting_app_id: int = betterproto.uint32_field(2)


class AckChatMessageNotification(UnifiedMessage, um_name="ChatRoom.AckChatMessage"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    timestamp: int = betterproto.uint32_field(3)


class CreateInviteLinkRequest(UnifiedMessage, um_name="ChatRoom.CreateInviteLink"):
    chat_group_id: int = betterproto.uint64_field(1)
    seconds_valid: int = betterproto.uint32_field(2)
    chat_id: int = betterproto.uint64_field(3)


class CreateInviteLinkResponse(UnifiedMessage, um_name="ChatRoom.CreateInviteLink"):
    invite_code: str = betterproto.string_field(1)
    seconds_valid: int = betterproto.uint32_field(2)


class GetInviteLinkInfoRequest(UnifiedMessage, um_name="ChatRoom.GetInviteLinkInfo"):
    invite_code: str = betterproto.string_field(1)


class GetInviteLinkInfoResponse(UnifiedMessage, um_name="ChatRoom.GetInviteLinkInfo"):
    steamid_sender: int = betterproto.fixed64_field(3)
    time_expires: int = betterproto.uint32_field(4)
    chat_id: int = betterproto.uint64_field(6)
    group_summary: "GetChatRoomGroupSummaryResponse" = betterproto.message_field(8)
    user_chat_group_state: "ChatRoomGroupState" = betterproto.message_field(9)
    time_kick_expire: int = betterproto.uint32_field(10)
    banned: bool = betterproto.bool_field(11)


class GetInviteInfoRequest(UnifiedMessage, um_name="ChatRoom.GetInviteInfo"):
    steamid_invitee: int = betterproto.fixed64_field(1)
    chat_group_id: int = betterproto.uint64_field(2)
    chat_id: int = betterproto.uint64_field(3)
    invite_code: str = betterproto.string_field(4)


class GetInviteInfoResponse(UnifiedMessage, um_name="ChatRoom.GetInviteInfo"):
    group_summary: "GetChatRoomGroupSummaryResponse" = betterproto.message_field(1)
    time_kick_expire: int = betterproto.uint32_field(2)
    banned: bool = betterproto.bool_field(3)


class GetInviteLinksForGroupRequest(UnifiedMessage, um_name="ChatRoom.GetInviteLinksForGroup"):
    chat_group_id: int = betterproto.uint64_field(1)


class GetInviteLinksForGroupResponse(UnifiedMessage, um_name="ChatRoom.GetInviteLinksForGroup"):
    invite_links: "list[GetInviteLinksForGroupResponseLinkInfo]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetInviteLinksForGroupResponseLinkInfo(betterproto.Message):
    invite_code: str = betterproto.string_field(1)
    steamid_creator: int = betterproto.fixed64_field(2)
    time_expires: int = betterproto.uint32_field(3)
    chat_id: int = betterproto.uint64_field(4)


class GetBanListRequest(UnifiedMessage, um_name="ChatRoom.GetBanList"):
    chat_group_id: int = betterproto.uint64_field(1)


class GetBanListResponse(UnifiedMessage, um_name="ChatRoom.GetBanList"):
    bans: "list[GetBanListResponseBanInfo]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetBanListResponseBanInfo(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    accountid_actor: int = betterproto.uint32_field(2)
    time_banned: int = betterproto.uint32_field(3)
    ban_reason: str = betterproto.string_field(4)


class GetInviteListRequest(UnifiedMessage, um_name="ChatRoom.GetInviteList"):
    chat_group_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class GroupInvite(betterproto.Message):
    accountid: int = betterproto.uint32_field(1)
    accountid_actor: int = betterproto.uint32_field(2)
    time_invited: int = betterproto.uint32_field(3)


class GetInviteListResponse(UnifiedMessage, um_name="ChatRoom.GetInviteList"):
    invites: "list[GroupInvite]" = betterproto.message_field(1)


class DeleteInviteLinkRequest(UnifiedMessage, um_name="ChatRoom.DeleteInviteLink"):
    chat_group_id: int = betterproto.uint64_field(1)
    invite_code: str = betterproto.string_field(2)


class DeleteInviteLinkResponse(UnifiedMessage, um_name="ChatRoom.DeleteInviteLink"):
    pass


class SetSessionActiveChatRoomGroupsRequest(UnifiedMessage, um_name="ChatRoom.SetSessionActiveChatRoomGroups"):
    chat_group_ids: list[int] = betterproto.uint64_field(1)
    chat_groups_data_requested: list[int] = betterproto.uint64_field(2)
    virtualize_members_threshold: int = betterproto.int32_field(3)


class SetSessionActiveChatRoomGroupsResponse(UnifiedMessage, um_name="ChatRoom.SetSessionActiveChatRoomGroups"):
    chat_states: "list[GroupState]" = betterproto.message_field(1)
    virtualize_members_chat_group_ids: list[int] = betterproto.uint64_field(2)


class SetUserChatGroupPreferencesRequest(UnifiedMessage, um_name="ChatRoom.SetUserChatGroupPreferences"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_group_preferences: "SetUserChatGroupPreferencesRequestChatGroupPreferences" = betterproto.message_field(2)
    chat_room_preferences: "list[SetUserChatGroupPreferencesRequestChatRoomPreferences]" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class SetUserChatGroupPreferencesRequestChatGroupPreferences(betterproto.Message):
    desktop_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(1)
    mobile_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(2)
    unread_indicator_muted: bool = betterproto.bool_field(3)


@dataclass(eq=False, repr=False)
class SetUserChatGroupPreferencesRequestChatRoomPreferences(betterproto.Message):
    chat_id: int = betterproto.uint64_field(1)
    desktop_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(2)
    mobile_notification_level: "EChatRoomNotificationLevel" = betterproto.enum_field(3)
    unread_indicator_muted: bool = betterproto.bool_field(4)


class SetUserChatGroupPreferencesResponse(UnifiedMessage, um_name="ChatRoom.SetUserChatGroupPreferences"):
    pass


class DeleteChatMessagesRequest(UnifiedMessage, um_name="ChatRoom.DeleteChatMessages"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    messages: "list[DeleteChatMessagesRequestMessage]" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class DeleteChatMessagesRequestMessage(betterproto.Message):
    server_timestamp: int = betterproto.uint32_field(1)
    ordinal: int = betterproto.uint32_field(2)


class DeleteChatMessagesResponse(UnifiedMessage, um_name="ChatRoom.DeleteChatMessages"):
    pass


class UpdateMemberListViewNotification(UnifiedMessage, um_name="ChatRoom.UpdateMemberListView"):
    chat_group_id: int = betterproto.uint64_field(1)
    view_id: int = betterproto.uint64_field(2)
    start: int = betterproto.int32_field(3)
    end: int = betterproto.int32_field(4)
    client_changenumber: int = betterproto.int32_field(5)
    delete_view: bool = betterproto.bool_field(6)
    persona_subscribe_accountids: list[int] = betterproto.int32_field(7)
    persona_unsubscribe_accountids: list[int] = betterproto.int32_field(8)


class SearchMembersRequest(UnifiedMessage, um_name="ChatRoom.SearchMembers"):
    chat_group_id: int = betterproto.uint64_field(1)
    search_id: int = betterproto.uint64_field(2)
    search_text: str = betterproto.string_field(3)
    max_results: int = betterproto.int32_field(4)


class SearchMembersResponse(UnifiedMessage, um_name="ChatRoom.SearchMembers"):
    matching_members: "list[SearchMembersResponseMemberMatch]" = betterproto.message_field(1)
    status_flags: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class SearchMembersResponseMemberMatch(betterproto.Message):
    accountid: int = betterproto.int32_field(1)
    persona: CMsgClientPersonaStateFriend = betterproto.message_field(2)


class UpdateMessageReactionRequest(UnifiedMessage, um_name="ChatRoom.UpdateMessageReaction"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    server_timestamp: int = betterproto.uint32_field(3)
    ordinal: int = betterproto.uint32_field(4)
    reaction_type: "EChatRoomMessageReactionType" = betterproto.enum_field(5)
    reaction: str = betterproto.string_field(6)
    is_add: bool = betterproto.bool_field(7)


class UpdateMessageReactionResponse(UnifiedMessage, um_name="ChatRoom.UpdateMessageReaction"):
    num_reactors: int = betterproto.uint32_field(1)


class GetMessageReactionReactorsRequest(UnifiedMessage, um_name="ChatRoom.GetMessageReactionReactors"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    server_timestamp: int = betterproto.uint32_field(3)
    ordinal: int = betterproto.uint32_field(4)
    reaction_type: "EChatRoomMessageReactionType" = betterproto.enum_field(5)
    reaction: str = betterproto.string_field(6)
    limit: int = betterproto.uint32_field(7)


class GetMessageReactionReactorsResponse(UnifiedMessage, um_name="ChatRoom.GetMessageReactionReactors"):
    reactors: list[int] = betterproto.uint32_field(1)


class GetClanChatRoomInfoRequest(UnifiedMessage, um_name="ClanChatRooms.GetClanChatRoomInfo"):
    steamid: int = betterproto.fixed64_field(1)
    autocreate: bool = betterproto.bool_field(2)


class GetClanChatRoomInfoResponse(UnifiedMessage, um_name="ClanChatRooms.GetClanChatRoomInfo"):
    chat_group_summary: "GetChatRoomGroupSummaryResponse" = betterproto.message_field(1)


class SetClanChatRoomPrivateRequest(UnifiedMessage, um_name="ClanChatRooms.SetClanChatRoomPrivate"):
    steamid: int = betterproto.fixed64_field(1)
    chat_room_private: bool = betterproto.bool_field(2)


class SetClanChatRoomPrivateResponse(UnifiedMessage, um_name="ClanChatRooms.SetClanChatRoomPrivate"):
    chat_room_private: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class Mentions(betterproto.Message):
    mention_all: bool = betterproto.bool_field(1)
    mention_here: bool = betterproto.bool_field(2)
    ids: list[int] = betterproto.uint32_field(3)


class IncomingChatMessageNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyIncomingChatMessage"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    steamid_sender: int = betterproto.fixed64_field(3)
    message: str = betterproto.string_field(4)
    timestamp: int = betterproto.uint32_field(5)
    mentions: "Mentions" = betterproto.message_field(6)
    ordinal: int = betterproto.uint32_field(7)
    server_message: "ServerMessage" = betterproto.message_field(8)
    message_no_bbcode: str = betterproto.string_field(9)
    chat_name: str = betterproto.string_field(10)


class ChatMessageModifiedNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyChatMessageModified"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    messages: "list[ChatMessageModifiedNotificationChatMessage]" = betterproto.message_field(3)


@dataclass(eq=False, repr=False)
class ChatMessageModifiedNotificationChatMessage(betterproto.Message):
    server_timestamp: int = betterproto.uint32_field(1)
    ordinal: int = betterproto.uint32_field(2)
    deleted: bool = betterproto.bool_field(3)


class MemberStateChangeNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyMemberStateChange"):
    chat_group_id: int = betterproto.uint64_field(1)
    member: "Member" = betterproto.message_field(2)
    change: "EChatRoomMemberStateChange" = betterproto.enum_field(3)


class ChatRoomHeaderStateNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyChatRoomHeaderStateChange"):
    header_state: "GroupHeaderState" = betterproto.message_field(1)


class ChatRoomGroupRoomsChangeNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyChatRoomGroupRoomsChange"):
    chat_group_id: int = betterproto.uint64_field(1)
    default_chat_id: int = betterproto.uint64_field(2)
    chat_rooms: "list[State]" = betterproto.message_field(3)


class NotifyShouldRejoinChatRoomVoiceChatNotification(
    UnifiedMessage, um_name="ChatRoomClient.NotifyShouldRejoinChatRoomVoiceChat"
):
    chat_id: int = betterproto.uint64_field(1)
    chat_group_id: int = betterproto.uint64_field(2)


class NotifyChatGroupUserStateChangedNotification(
    UnifiedMessage, um_name="ChatRoomClient.NotifyChatGroupUserStateChanged"
):
    chat_group_id: int = betterproto.uint64_field(1)
    user_chat_group_state: "ChatRoomGroupState" = betterproto.message_field(2)
    group_summary: "GetChatRoomGroupSummaryResponse" = betterproto.message_field(3)
    user_action: "EChatRoomMemberStateChange" = betterproto.enum_field(4)


class NotifyChatRoomDisconnectNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyChatRoomDisconnect"):
    chat_group_ids: list[int] = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class MemberListView(betterproto.Message):
    start: int = betterproto.int32_field(3)
    end: int = betterproto.int32_field(4)
    total_count: int = betterproto.int32_field(5)
    client_changenumber: int = betterproto.int32_field(6)
    server_changenumber: int = betterproto.int32_field(7)


@dataclass(eq=False, repr=False)
class MemberSummaryCounts(betterproto.Message):
    ingame: int = betterproto.int32_field(1)
    online: int = betterproto.int32_field(2)
    offline: int = betterproto.int32_field(3)


class MemberListViewUpdatedNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyMemberListViewUpdated"):
    chat_group_id: int = betterproto.uint64_field(1)
    view_id: int = betterproto.uint64_field(2)
    view: "MemberListView" = betterproto.message_field(3)
    members: "list[MemberListViewUpdatedNotificationMemberListViewEntry]" = betterproto.message_field(4)
    status_flags: int = betterproto.uint32_field(5)
    member_summary: "MemberSummaryCounts" = betterproto.message_field(6)
    subscribed_personas: list[CMsgClientPersonaStateFriend] = betterproto.message_field(7)


@dataclass(eq=False, repr=False)
class MemberListViewUpdatedNotificationMemberListViewEntry(betterproto.Message):
    rank: int = betterproto.int32_field(1)
    accountid: int = betterproto.uint32_field(2)
    persona: CMsgClientPersonaStateFriend = betterproto.message_field(3)


class MessageReactionNotification(UnifiedMessage, um_name="ChatRoomClient.NotifyMessageReaction"):
    chat_group_id: int = betterproto.uint64_field(1)
    chat_id: int = betterproto.uint64_field(2)
    server_timestamp: int = betterproto.uint32_field(3)
    ordinal: int = betterproto.uint32_field(4)
    reactor: int = betterproto.fixed64_field(5)
    reaction_type: "EChatRoomMessageReactionType" = betterproto.enum_field(6)
    reaction: str = betterproto.string_field(7)
    is_add: bool = betterproto.bool_field(8)


@dataclass(eq=False, repr=False)
class UsabilityClientUsabilityMetricsNotification(betterproto.Message):
    metrics_run_id: int = betterproto.uint32_field(1)
    client_build: int = betterproto.uint32_field(2)
    metrics_version: int = betterproto.uint32_field(3)
    in_web: bool = betterproto.bool_field(4)
    settings: "UsabilityClientUsabilityMetricsNotificationSettings" = betterproto.message_field(10)
    voice_settings: "UsabilityClientUsabilityMetricsNotificationVoiceSettings" = betterproto.message_field(11)
    ui_state: "UsabilityClientUsabilityMetricsNotificationUiState" = betterproto.message_field(12)
    metrics: "UsabilityClientUsabilityMetricsNotificationMetrics" = betterproto.message_field(13)


@dataclass(eq=False, repr=False)
class UsabilityClientUsabilityMetricsNotificationSettings(betterproto.Message):
    notifications_show_ingame: bool = betterproto.bool_field(1)
    notifications_show_online: bool = betterproto.bool_field(2)
    notifications_show_message: bool = betterproto.bool_field(3)
    notifications_events_and_announcements: bool = betterproto.bool_field(4)
    sounds_play_ingame: bool = betterproto.bool_field(5)
    sounds_play_online: bool = betterproto.bool_field(6)
    sounds_play_message: bool = betterproto.bool_field(7)
    sounds_events_and_announcements: bool = betterproto.bool_field(8)
    always_new_chat_window: bool = betterproto.bool_field(9)
    force_alphabetic_friend_sorting: bool = betterproto.bool_field(10)
    chat_flash_mode: int = betterproto.int32_field(11)
    remember_open_chats: bool = betterproto.bool_field(12)
    compact_quick_access: bool = betterproto.bool_field(13)
    compact_friends_list: bool = betterproto.bool_field(14)
    notifications_show_chat_room_notification: bool = betterproto.bool_field(15)
    sounds_play_chat_room_notification: bool = betterproto.bool_field(16)
    hide_offline_friends_in_tag_groups: bool = betterproto.bool_field(17)
    hide_categorized_friends: bool = betterproto.bool_field(18)
    categorize_in_game_friends_by_game: bool = betterproto.bool_field(19)
    chat_font_size: int = betterproto.int32_field(20)
    use24_hour_clock: bool = betterproto.bool_field(21)
    do_not_disturb_mode: bool = betterproto.bool_field(22)
    disable_embed_inlining: bool = betterproto.bool_field(23)
    sign_into_friends: bool = betterproto.bool_field(24)
    animated_avatars: bool = betterproto.bool_field(25)


@dataclass(eq=False, repr=False)
class UsabilityClientUsabilityMetricsNotificationVoiceSettings(betterproto.Message):
    voice_input_gain: float = betterproto.float_field(1)
    voice_output_gain: float = betterproto.float_field(2)
    noise_gate_level: int = betterproto.int32_field(3)
    voice_use_echo_cancellation: bool = betterproto.bool_field(4)
    voice_use_noise_cancellation: bool = betterproto.bool_field(5)
    voice_use_auto_gain_control: bool = betterproto.bool_field(6)
    selected_non_default_mic: bool = betterproto.bool_field(7)
    selected_non_default_output: bool = betterproto.bool_field(8)
    push_to_talk_enabled: bool = betterproto.bool_field(9)
    push_to_mute_enabled: bool = betterproto.bool_field(10)
    play_ptt_sounds: bool = betterproto.bool_field(11)


@dataclass(eq=False, repr=False)
class UsabilityClientUsabilityMetricsNotificationUiState(betterproto.Message):
    friends_list_height: int = betterproto.int32_field(1)
    friends_list_width: int = betterproto.int32_field(2)
    friends_list_docked: bool = betterproto.bool_field(3)
    friends_list_collapsed: bool = betterproto.bool_field(4)
    friends_list_group_chats_height: int = betterproto.int32_field(5)
    friends_list_visible: bool = betterproto.bool_field(6)
    chat_popups_opened: int = betterproto.int32_field(7)
    group_chat_tabs_opened: int = betterproto.int32_field(8)
    friend_chat_tabs_opened: int = betterproto.int32_field(9)
    chat_window_width: int = betterproto.int32_field(10)
    chat_window_height: int = betterproto.int32_field(11)
    category_collapse: "UsabilityClientUsabilityMetricsNotificationUiStateCategoryCollapseState" = (
        betterproto.message_field(12)
    )
    group_chat_left_col_collapsed: int = betterproto.int32_field(13)
    group_chat_right_col_collapsed: int = betterproto.int32_field(14)
    in_one_on_one_voice_chat: bool = betterproto.bool_field(15)
    in_group_voice_chat: bool = betterproto.bool_field(16)


@dataclass(eq=False, repr=False)
class UsabilityClientUsabilityMetricsNotificationUiStateCategoryCollapseState(betterproto.Message):
    in_game_collapsed: bool = betterproto.bool_field(1)
    online_collapsed: bool = betterproto.bool_field(2)
    offline_collapsed: bool = betterproto.bool_field(3)
    game_groups_collapsed: int = betterproto.int32_field(4)
    categories_collapsed: int = betterproto.int32_field(5)


@dataclass(eq=False, repr=False)
class UsabilityClientUsabilityMetricsNotificationMetrics(betterproto.Message):
    friends_count: int = betterproto.int32_field(1)
    friends_category_count: int = betterproto.int32_field(2)
    friends_categorized_count: int = betterproto.int32_field(3)
    friends_online_count: int = betterproto.int32_field(4)
    friends_in_game_count: int = betterproto.int32_field(5)
    friends_in_game_singleton_count: int = betterproto.int32_field(6)
    game_group_count: int = betterproto.int32_field(7)
    friends_favorite_count: int = betterproto.int32_field(8)
    group_chat_count: int = betterproto.int32_field(9)
    group_chat_favorite_count: int = betterproto.int32_field(10)


@dataclass(eq=False, repr=False)
class UsabilityRequestClientUsabilityMetricsNotification(betterproto.Message):
    metrics_run_id: int = betterproto.uint32_field(1)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: service_quest.proto
# plugin: python-betterproto
# Last updated 21/12/2022

from dataclasses import dataclass

import betterproto

from .msg import UnifiedMessage


class ActivateProfileModifierItemRequest(UnifiedMessage, um_name="Quest.ActivateProfileModifierItem"):
    appid: int = betterproto.uint32_field(1)
    communityitemid: int = betterproto.uint64_field(2)
    activate: bool = betterproto.bool_field(3)


class ActivateProfileModifierItemResponse(UnifiedMessage, um_name="Quest.ActivateProfileModifierItem"):
    pass


@dataclass(eq=False, repr=False)
class CommunityItem(betterproto.Message):
    communityitemid: int = betterproto.uint64_field(1)
    item_type: int = betterproto.uint32_field(2)
    appid: int = betterproto.uint32_field(3)
    owner: int = betterproto.uint32_field(4)
    attributes: "list[CommunityItemAttribute]" = betterproto.message_field(5)
    used: bool = betterproto.bool_field(6)
    owner_origin: int = betterproto.uint32_field(7)
    amount: int = betterproto.int64_field(8)


@dataclass(eq=False, repr=False)
class CommunityItemAttribute(betterproto.Message):
    attributeid: int = betterproto.uint32_field(1)
    value: int = betterproto.uint64_field(2)


class GetCommunityInventoryRequest(UnifiedMessage, um_name="Quest.GetCommunityInventory"):
    filter_appids: list[int] = betterproto.uint32_field(1)


class GetCommunityInventoryResponse(UnifiedMessage, um_name="Quest.GetCommunityInventory"):
    items: "list[CommunityItem]" = betterproto.message_field(1)


class GetCommunityItemDefinitionsRequest(UnifiedMessage, um_name="Quest.GetCommunityItemDefinitions"):
    appid: int = betterproto.uint32_field(1)
    item_type: int = betterproto.uint32_field(3)
    language: str = betterproto.string_field(4)
    broadcast_channel_id: int = betterproto.uint64_field(5)
    keyvalues_as_json: bool = betterproto.bool_field(6)


class GetCommunityItemDefinitionsResponse(UnifiedMessage, um_name="Quest.GetCommunityItemDefinitions"):
    item_definitions: "list[GetCommunityItemDefinitionsResponseItemDefinition]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetCommunityItemDefinitionsResponseItemDefinition(betterproto.Message):
    item_type: int = betterproto.uint32_field(1)
    appid: int = betterproto.uint32_field(2)
    item_name: str = betterproto.string_field(3)
    item_title: str = betterproto.string_field(4)
    item_description: str = betterproto.string_field(5)
    item_image_small: str = betterproto.string_field(6)
    item_image_large: str = betterproto.string_field(7)
    item_key_values: str = betterproto.string_field(8)
    item_series: int = betterproto.uint32_field(9)
    item_class: int = betterproto.uint32_field(10)
    editor_accountid: int = betterproto.uint32_field(11)
    active: bool = betterproto.bool_field(12)
    item_image_composed: str = betterproto.string_field(13)
    item_image_composed_foil: str = betterproto.string_field(14)
    deleted: bool = betterproto.bool_field(15)
    item_last_changed: int = betterproto.uint32_field(16)
    broadcast_channel_id: int = betterproto.uint64_field(17)
    item_movie_webm: str = betterproto.string_field(18)
    item_movie_mp4: str = betterproto.string_field(19)
    item_movie_webm_small: str = betterproto.string_field(20)
    item_movie_mp4_small: str = betterproto.string_field(21)


class GetNumTradingCardsEarnedRequest(UnifiedMessage, um_name="Quest.GetNumTradingCardsEarned"):
    timestamp_start: int = betterproto.uint32_field(1)
    timestamp_end: int = betterproto.uint32_field(2)


class GetNumTradingCardsEarnedResponse(UnifiedMessage, um_name="Quest.GetNumTradingCardsEarned"):
    num_trading_cards: int = betterproto.uint32_field(1)


class SetVirtualItemRewardDefinitionRequest(UnifiedMessage, um_name="Quest.SetVirtualItemRewardDefinition"):
    eventid: int = betterproto.int32_field(1)
    itemsdefs: "list[CVirtualItemRewardDefinition]" = betterproto.message_field(2)
    action: int = betterproto.int32_field(3)


class SetVirtualItemRewardDefinitionResponse(UnifiedMessage, um_name="Quest.SetVirtualItemRewardDefinition"):
    pass


class VirtualItemRewardDefinitionRequest(UnifiedMessage, um_name="Quest.VirtualItemRewardDefinition"):
    eventid: int = betterproto.int32_field(1)
    include_inactive: bool = betterproto.bool_field(2)


class VirtualItemRewardDefinitionResponse(UnifiedMessage, um_name="Quest.VirtualItemRewardDefinition"):
    rewards: "list[CVirtualItemRewardDefinition]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class CVirtualItemRewardDefinition(betterproto.Message):
    eventid: int = betterproto.int32_field(1)
    item_bucket: int = betterproto.uint32_field(2)
    appid: int = betterproto.uint32_field(3)
    active: bool = betterproto.bool_field(4)
    rarity: int = betterproto.uint32_field(5)
    package_to_grant: int = betterproto.uint32_field(6)
    game_item_id: int = betterproto.fixed64_field(7)
    community_item_class: int = betterproto.int32_field(8)
    community_item_type: int = betterproto.uint32_field(9)
    loyalty_point_type: int = betterproto.uint32_field(10)
    amount: int = betterproto.int64_field(11)
    rtime_time_active: int = betterproto.uint32_field(12)
    loyalty_reward_defid: int = betterproto.uint32_field(13)
    user_badge_to_grant: int = betterproto.uint32_field(14)
    user_badge_level: int = betterproto.uint32_field(15)
    virtual_item_def_id: int = betterproto.uint32_field(16)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import types
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Sequence
from inspect import get_annotations
from typing import (
    TYPE_CHECKING,
    Any,
    Final,
    ForwardRef,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    final,
    get_args,
    get_origin,
    runtime_checkable,
)

from typing_extensions import Never, Self, get_original_bases

from ... import utils
from ...abc import PartialUser
from ...app import App, PartialApp
from ...channel import Channel
from ...chat import ChatMessage, Member, PartialMember
from ...clan import Clan
from ...enums import Type
from ...errors import HTTPException, InvalidID, WSException
from ...group import Group
from ...id import id64_from_url, parse_id64
from ...user import ClientUser, User
from .errors import BadArgument, MissingRequiredArgument

if TYPE_CHECKING:
    from steam.ext import commands

    from ...friend import Friend
    from .bot import Bot
    from .commands import MaybeCommandT
    from .context import Context

__all__ = (
    "converter",
    "Converter",
    "UserConverter",
    "ChannelConverter",
    "ClanConverter",
    "GroupConverter",
    "AppConverter",
    "Default",
    "DefaultAuthor",
    "DefaultChannel",
    "DefaultClan",
    "DefaultGroup",
    "DefaultApp",
    "Greedy",
)

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
BotT = TypeVar("BotT", bound="Bot", covariant=True)
Converters: TypeAlias = "ConverterBase[Any] | BasicConverter[Any] | Callable[[str], Any]"


class BasicConverter(Protocol[T]):
    converter_for: type[T]

    def __call__(self, arg: str, /) -> T: ...


CONVERTERS = defaultdict[type, list[Converters]](list)


def converter(func: Callable[[str], T], /) -> BasicConverter[T]:
    """The recommended way to mark a function converter as such.

    Note
    ----
    All of the converters marked with this decorator or derived from :class:`Converter` can be accessed via
    :attr:`~commands.Bot.converters`.

    Examples
    --------
    .. code:: python

        @commands.converter
        def command_converter(argument: str) -> commands.Command:  # this is the type hint used
            return bot.get_command(argument)


        # then later
        @bot.command
        async def source(ctx, command: commands.Command):  # this then calls command_converter on invocation.
            ...


    Attributes
    -----------
    converter_for: T
        The class that the converter can be type-hinted to to.
    """

    annotations = get_annotations(func)
    converter_for = annotations["return"]
    func = cast(BasicConverter[T], func)
    func.converter_for = converter_for
    CONVERTERS[converter_for].append(func)
    return func


@runtime_checkable
class ConverterBase(Protocol[T_co]):
    # this is the base class we use for isinstance checks, don't actually this
    @abstractmethod
    async def convert(self, ctx: commands.Context[Any], argument: str, /) -> T_co:
        """An abstract method all converters must derive.

        Parameters
        ----------
        ctx
            The context for the invocation.
        argument
            The argument that is passed from the argument parser.

        Returns
        -------
        An object, should be of the same type of :attr:`converter_for`.
        """
        raise NotImplementedError("Derived classes must implement this")


class Converter(ConverterBase[T_co], ABC):
    """A custom :class:`typing.Protocol` from which converters can be derived.

    Note
    ----
    All of the converters derived from :class:`Converter` or marked with the :func:`converter_for` decorator can
    be accessed via :attr:`~commands.Bot.converters`.

    Some custom dataclasses from this library can be type-hinted without the need for a custom converter:

        - :class:`~steam.User`.
        - :class:`~steam.abc.Channel`
        - :class:`~steam.Clan`
        - :class:`~steam.Group`
        - :class:`~steam.App`

    Examples
    --------

    Builtin:

    .. code:: python

        @bot.command
        async def command(
            ctx, user: steam.User
        ): ...  # this will tell the parser to convert user from a str to a steam.User object.


        # invoked as
        # !command 76561198248053954
        # or !command Gobot1234

    A custom converter:

    .. code:: python

        class MediaConverter(commands.Converter[steam.Media]):  # the annotation to typehint to
            async def convert(self, ctx: commands.Context, argument: str) -> steam.Media:
                async with aiohttp.ClientSession() as session:
                    async with session.get(argument) as r:
                        media_bytes = BytesIO(await r.read())
                try:
                    return steam.Media(media_bytes)
                except (TypeError, ValueError):  # failed to convert to an media
                    raise commands.BadArgument("Cannot convert media") from None


        # then later
        @bot.command
        async def set_avatar(ctx: commands.Context, avatar: steam.Media) -> None:
            await bot.user.edit(avatar=avatar)
            await ctx.send("👌")


        # invoked as
        # !set_avatar https://my_image_url.com
    """

    converter_for: type[T_co]

    @classmethod
    def register(cls, command: MaybeCommandT) -> MaybeCommandT:
        """Register a converter to a specific command.

        Examples
        --------
        .. code:: python

            class CustomUserConverter(commands.Converter[steam.User]):
                async def convert(self, ctx: commands.Context, argument: str) -> steam.User: ...


            @bot.command
            @CustomUserConverter.register
            async def is_cool(ctx, user: steam.User): ...

        In this example ``is_cool``'s user parameter would be registered to the ``CustomUserConverter`` rather than
        the global :class:`UserConverter`.
        """
        from .commands import Command

        try:
            (
                command.special_converters
                if isinstance(command, Command)
                else cast(list[Converters], command.__special_converters__)  # type: ignore
            ).append(cls)
        except AttributeError:
            command.__special_converters__ = [cls]  # type: ignore
        return command

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        try:
            converter_for = get_args(get_original_bases(cls)[-1])[0]  # T_co.__value__
        except IndexError:
            raise TypeError("Converters should subclass commands.Converter using __class_getitem__")
        else:
            if isinstance(converter_for, ForwardRef):
                raise NameError(f"name {converter_for.__forward_arg__!r} is not defined") from None
            cls.converter_for = converter_for
            """The class that the converter can be type-hinted to to."""
            CONVERTERS[converter_for].append(cls)


class PartialUserConverter(Converter[PartialUser]):
    async def convert(self, ctx: Context, argument: str) -> PartialUser:
        try:
            return ctx._state.get_partial_user(argument)
        except InvalidID:
            if (
                argument.startswith("@")
                and isinstance(ctx.message, ChatMessage)
                and (
                    (
                        mention := utils.find(
                            lambda tag: (
                                tag.name == "mention" and tag.inner == argument and tag.position[0] >= ctx.lex.position
                            ),
                            ctx.message.content.tags,
                        )
                    )
                    is not None
                )
            ):
                return ctx._state.get_partial_user(mention.attributes[""])
            if (user := utils.get(ctx.bot.users, name=argument)) is not None:
                return ctx._state.get_partial_user(user.id64)
            if (id64 := await id64_from_url(argument, session=ctx._state.http._session)) is not None:
                return ctx._state.get_partial_user(id64)

        raise BadArgument(f'Failed to convert "{argument}" to a Steam user')


class UserConverter(PartialUserConverter, Converter[User]):
    """The converter that is used when the type-hint passed is :class:`~steam.User`.

    Lookup is in the order of:
        - Steam ID
        - Mentions
        - Name
        - URLs
    """

    async def convert(self, ctx: Context, argument: str) -> User | ClientUser:
        partial_user = await super().convert(ctx, argument)
        try:
            user = await ctx._state._maybe_user(partial_user.id64)
            if not isinstance(user, (User, ClientUser)):
                raise BadArgument(f'Failed to convert "{argument}" to a Steam user') from None
            return user
        except WSException:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam user') from None


class PartialMemberConverter(PartialUserConverter, Converter[PartialMember]):
    async def convert(self, ctx: Context, argument: str) -> PartialMember:
        partial_user = await super().convert(ctx, argument)
        if not (ctx.clan or ctx.group):
            raise BadArgument("Cannot convert to a partial member without a clan or group")
        return ctx._chat_group._get_partial_member(partial_user.id)


class MemberConverter(PartialMemberConverter, Converter[Member]):
    async def convert(self, ctx: Context, argument: str) -> Member:
        partial_member = await super().convert(ctx, argument)
        try:
            user = ctx._chat_group._maybe_member(partial_member.id)
            if not isinstance(user, Member):
                raise TypeError
            return user
        except (WSException, TypeError):
            raise BadArgument(f'Failed to convert "{argument}" to a Steam user') from None


class ChannelConverter(Converter[Channel[Any, Any, Any]]):
    """The converter that is used when the type-hint passed is :class:`~steam.Channel`.

    Lookup is in the order of:
        - ID
        - Name
    """

    async def convert(self, ctx: Context, argument: str) -> Channel[Any]:
        channel = utils.find(
            (lambda c: c.id == int(argument)) if argument.isdigit() else lambda c: c.name == argument,
            ctx._chat_group.channels,
        )
        if channel is None:
            raise BadArgument(f'Failed to convert "{argument}" to a channel')
        return channel


class ClanConverter(Converter[Clan]):
    """The converter that is used when the type-hint passed is :class:`~steam.Clan`.

    Lookup is in the order of:
        - Steam ID
        - Name
        - URLs
    """

    async def convert(self, ctx: Context, argument: str) -> Clan:
        try:
            clan = await ctx.bot.fetch_clan(parse_id64(argument, type=Type.Clan))
        except (InvalidID, HTTPException):
            clan = utils.find(lambda c: c.name == argument, ctx.bot.clans)
            if clan is None:
                id64 = await id64_from_url(argument, session=ctx._state.http._session)
                if id64 is None:
                    raise BadArgument(f'Failed to convert "{argument}" to a Steam clan')
                return await ctx.bot.fetch_clan(id64)
        except WSException:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam clan')
        return clan


class GroupConverter(Converter[Group]):
    """The converter that is used when the type-hint passed is :class:`~steam.Group`.

    Lookup is in the order of:
        - ID
        - Name
    """

    async def convert(self, ctx: Context, argument: str) -> Group:
        try:
            group = ctx.bot.get_group(parse_id64(argument, type=Type.Chat))
        except InvalidID:
            group = utils.find(lambda c: c.name == argument, ctx.bot.groups)
        if group is None:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam group')
        return group


class AppConverter(Converter[App]):
    """The converter that is used when the type-hint passed is :class:`~steam.App`.

    If the param is a digit it is assumed that the argument is the :attr:`App.id` else it is assumed it is the
    :attr:`App.name`.
    """

    async def convert(self, ctx: Context, argument: str) -> App[str]:
        entries = await ctx._state.http.get_app_suggestions(argument)
        try:
            return next(App(id=int(app["id"]), name=app["name"]) for app in entries)
        except StopIteration:
            raise BadArgument(f'Failed to convert "{argument}" to a Steam app') from None


class PartialAppConverter(AppConverter, Converter[PartialApp[str]]):
    """The converter that is used when the type-hint passed is :class:`~steam.App`.

    If the param is a digit it is assumed that the argument is the :attr:`App.id` else it is assumed it is the
    :attr:`App.name`.
    """

    async def convert(self, ctx: Context, argument: str) -> PartialApp[str]:
        app = await super().convert(ctx, argument)
        return PartialApp(ctx._state, id=app.id, name=app.name)


@runtime_checkable
class Default(Protocol, Any if TYPE_CHECKING else object):
    """A custom way to specify a default values for commands.

    Examples
    --------
    Builtin:

    .. code:: python

        @bot.command()
        async def info(ctx, user=DefaultAuthor): ...  # if no user is passed it will be ctx.author

    A custom default:

    .. code:: python

        class CurrentCommand(commands.Default):
            async def default(self, ctx: commands.Context) -> commands.Command:
                return ctx.command  # return the current command


        # then later
        @bot.command
        async def source(ctx: commands.Context, command: commands.Command = CurrentCommand): ...  # command would now be source


        # this could also be mixed in with a converter to convert a string to a command.
    """

    @abstractmethod
    async def default(self, ctx: commands.Context) -> object:
        raise NotImplementedError("Derived classes must to implement this")


class DefaultAuthor(Default):
    """Returns the :attr:`.Context.author`"""

    async def default(self, ctx: Context) -> User | ClientUser | Friend | Member | PartialMember | PartialUser:
        return ctx.author


class DefaultChannel(Default):
    """Returns the :attr:`.Context.channel`"""

    async def default(self, ctx: Context) -> Channel[Any]:
        return ctx.channel


class DefaultGroup(Default):
    """Returns the :attr:`.Context.group`"""

    async def default(self, ctx: Context) -> Group:
        if not ctx.group:
            assert ctx.current_param is not None
            raise MissingRequiredArgument(ctx.current_param)
        return ctx.group


class DefaultClan(Default):
    """Returns the :attr:`.Context.clan`"""

    async def default(self, ctx: Context) -> Clan:
        if not ctx.clan:
            assert ctx.current_param is not None
            raise MissingRequiredArgument(ctx.current_param)
        return ctx.clan


class DefaultApp(Default):
    """Returns the :attr:`.Context.author`'s :attr:`~steam.User.app`"""

    async def default(self, ctx: Context) -> PartialApp:
        if getattr(ctx.author, "app", None) is None:
            assert ctx.current_param is not None
            raise MissingRequiredArgument(ctx.current_param)
        return ctx.author.app  # type: ignore


class GreedyGenericAlias(types.GenericAlias):
    converter: Any

    def __getattribute__(self, name: str, /) -> Any:
        if name == "converter":
            return self.__args__[0]
        return super().__getattribute__(name)


@final
class Greedy(Sequence[T]):
    """
    A custom :class:`typing.Generic` that allows for special greedy command parsing behaviour. It signals to the command
    parser to consume as many arguments as it can until it silently errors reverts the last argument being read and
    then carries on normally.

    Greedy can be mixed in with any normally supported positional or keyword argument type any number of times.

    Example
    -------
    .. code:: python

        @bot.command()
        async def test(ctx, numbers: commands.Greedy[int], reason: str):
            await ctx.send(f"numbers: {numbers}, reason: {reason}")

    An invocation of ``"test 1 2 3 4 5 6 hello"`` would pass ``(1, 2, 3, 4, 5, 6)`` to ``numbers`` and ``"hello"`` to
    ``reason``.
    """

    converter: Final[type[T]] = Any  # unused
    """The converter the Greedy type holds."""

    def __new__(
        cls, *args: Any, **kwargs: Any
    ) -> Never:  # give a more helpful message than typing._BaseGenericAlias.__call__
        raise TypeError("Greedy cannot be instantiated directly, instead use Greedy[...]")

    def __class_getitem__(cls, converter: T | tuple[T]) -> Self:
        """The entry point for creating a Greedy type."""
        if isinstance(converter, tuple):
            try:
                (converter,) = converter
            except ValueError:
                raise TypeError("Cannot pass variadic arguments to Greedy")
        if isinstance(converter, ForwardRef):
            converter = eval(converter.__forward_code__)

        if not (callable(converter) or isinstance(converter, (Converter, str)) or get_origin(converter) is not None):
            raise TypeError(f"Greedy[...] expects a type or a Converter instance not {converter!r}")

        if (isinstance(converter, type) and issubclass(converter, str)) or converter is types.NoneType:
            raise TypeError(f"Greedy[{converter.__name__}] is invalid")  # type: ignore

        return cast(Self, GreedyGenericAlias(cls, (converter,)))



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import itertools
import re
import struct
from typing import TYPE_CHECKING, Final, Literal, overload

from ..._const import DOCS_BUILDING, MISSING, timeout
from ..._gc import Client as Client_
from ...app import CSGO
from ...enums import Type
from ...ext import commands
from ...id import ID, parse_id64
from ...utils import cached_property  # noqa: TC001
from .backpack import BaseInspectedItem, Paint, Sticker
from .enums import ItemOrigin, ItemQuality
from .models import ClientUser, Match, User
from .protobufs import cstrike
from .state import GCState  # noqa: TC001
from .utils import decode_sharecode

if TYPE_CHECKING:
    from ...ext import csgo
    from ...friend import Friend
    from ...types.id import Intable
    from ...types.user import IndividualID


__all__ = (
    "Client",
    "Bot",
)


class Client(Client_):
    """Represents a client connection that connects to Steam. This class is used to interact with the Steam API, CMs
    and the CSGO Game Coordinator.

    :class:`Client` is a subclass of :class:`steam.Client`, so whatever you can do with :class:`steam.Client` you can
    do with :class:`Client`.
    """

    _APP: Final = CSGO
    _ClientUserCls = ClientUser
    _state: GCState  # type: ignore  # PEP 705
    _GC_HEART_BEAT = 10.0
    if TYPE_CHECKING:

        @cached_property
        def user(self) -> ClientUser: ...

    @overload
    async def inspect_item(self, *, owner: IndividualID, asset_id: int, d: int) -> BaseInspectedItem: ...

    @overload
    async def inspect_item(self, *, market_id: int, asset_id: int, d: int) -> BaseInspectedItem: ...

    @overload
    async def inspect_item(self, *, url: str) -> BaseInspectedItem: ...

    async def inspect_item(
        self,
        *,
        owner: IndividualID | None = None,
        asset_id: int = 0,
        d: int = 0,
        market_id: int = 0,
        url: str = "",
    ) -> BaseInspectedItem:
        """Inspect an item.

        Parameters
        ----------
        owner
            The owner of the item.
        asset_id
            The asset id of the item.
        d
            The "D" number following the "D" character.
        market_id
            The id of the item on the steam community market.
        url
            The full inspect url to be parsed.
        """

        if url:
            search = re.search(r"[SM](\d+)A(\d+)D(\d+)$", url)
            if search is None:
                raise ValueError("Inspect url is invalid")

            owner = (
                ID[Literal[Type.Individual]](int(search[1]), type=Type.Individual)
                if search[0].startswith("S")
                else None
            )
            market_id = int(search[1]) if search[0].startswith("M") else 0
            asset_id = int(search[2])
            d = int(search[3])

        elif owner is None and market_id == 0:
            raise TypeError("Missing required keyword-only argument: 'owner' or 'market_id'")
        elif d == 0 or asset_id == 0:
            raise TypeError(f"Missing required keyword-only argument: {'asset_id' if d else 'd'}")

        future = self._state.ws.gc_wait_for(
            cstrike.Client2GcEconPreviewDataBlockResponse,
            check=lambda msg: msg.iteminfo.itemid == asset_id,
        )
        await self._state.ws.send_gc_message(
            cstrike.Client2GcEconPreviewDataBlockRequest(
                param_s=owner.id64 if owner else 0,
                param_a=asset_id,
                param_d=d,
                param_m=market_id,
            )
        )

        msg = await future

        item = msg.iteminfo
        # decode the wear
        packed_wear = struct.pack(">l", item.paintwear)
        (paint_wear,) = struct.unpack(">f", packed_wear)
        return BaseInspectedItem(
            id=item.itemid,
            def_index=item.defindex,
            paint=Paint(index=item.paintindex, wear=paint_wear, seed=item.paintseed),
            rarity=item.rarity,
            quality=ItemQuality.try_value(item.quality),
            kill_eater_score_type=item.killeaterscoretype,
            kill_eater_value=item.killeatervalue,
            custom_name=item.customname,
            stickers=[
                Sticker(
                    slot=sticker.slot,  # type: ignore
                    id=sticker.sticker_id,
                    wear=sticker.wear,
                    scale=sticker.scale,
                    rotation=sticker.rotation,
                    tint_id=sticker.tint_id,
                )
                for sticker in item.stickers
            ],
            inventory=item.inventory,
            origin=ItemOrigin.try_value(item.origin),
            quest_id=item.questid,
            drop_reason=item.dropreason,
            music_index=item.musicindex,
            ent_index=item.entindex,
        )

    @overload
    async def fetch_match(self, id: int, *, outcome_id: int, token: int) -> Match: ...

    @overload
    async def fetch_match(self, *, code: str) -> Match: ...

    async def fetch_match(
        self, id: int = MISSING, *, outcome_id: int = MISSING, token: int = MISSING, code: str = MISSING
    ) -> Match:
        """Fetch a match by its id or share code."""
        if code is not MISSING:
            id, outcome_id, token = decode_sharecode(code)

        future = self._state.ws.gc_wait_for(
            cstrike.MatchList,
            check=lambda msg: (
                msg.msgrequestid == cstrike.MatchListRequestFullGameInfo.MSG
                and bool(msg.matches)
                and msg.matches[0].matchid == id
            ),
        )
        await self._state.ws.send_gc_message(cstrike.MatchListRequestFullGameInfo(id, outcome_id, token))
        async with timeout(30):
            msg = await future

        (match,) = msg.matches
        players = {
            player.id: player
            for player in await self._state._maybe_users(
                dict.fromkeys(
                    map(
                        parse_id64,
                        itertools.chain.from_iterable(stat.reservation.account_ids for stat in match.roundstatsall),
                    )
                )
            )
        }
        return Match(self._state, match, players)

    async def fetch_friend_watch_info(self, *friends: Friend) -> cstrike.WatchInfoUsers:
        future = self._state.ws.gc_wait_for(cstrike.WatchInfoUsers)
        await self._state.ws.send_gc_message(
            cstrike.ClientRequestWatchInfoFriends(account_ids=[friend.id for friend in friends], request_id=1)
        )
        return await future

    if TYPE_CHECKING or DOCS_BUILDING:

        def get_user(self, id: Intable) -> User | None: ...

        async def fetch_user(self, id: Intable) -> User: ...

        async def on_gc_connect(self) -> None:
            """Called after the client receives the welcome message from the GC.

            Warning
            -------
            This is called every time we craft an item and disconnect so same warnings apply to
            :meth:`steam.Client.on_connect`
            """

        async def on_gc_disconnect(self) -> None:
            """Called after the client receives the goodbye message from the GC.

            Warning
            -------
            This is called every time we craft an item and disconnect so same warnings apply to
            :meth:`steam.Client.on_connect`
            """

        async def on_gc_ready(self) -> None:
            """Called after the client connects to the GC and has the :attr:`schema`, :meth:`Client.user.inventory` and
            set up and account info (:meth:`is_premium` and :attr:`backpack_slots`).

            Warning
            -------
            This is called every time we craft an item and disconnect so same warnings apply to
            :meth:`steam.Client.on_connect`
            """

        async def on_item_receive(self, item: csgo.BackpackItem) -> None:
            """Called when the client receives an item.

            Parameters
            ----------
            item
                The received item.
            """

        async def on_item_remove(self, item: csgo.BackpackItem) -> None:
            """Called when the client has an item removed from its backpack.

            Parameters
            ----------
            item
                The removed item.
            """

        async def on_item_update(self, before: csgo.BackpackItem, after: csgo.BackpackItem) -> None:
            """Called when the client has an item in its backpack updated.

            Parameters
            ----------
            before
                The item before being updated.
            after
                The item now.
            """


class Bot(commands.Bot, Client):
    """Represents a Steam bot.

    :class:`Bot` is a subclass of :class:`~steam.ext.commands.Bot`, so whatever you can do with
    :class:`~steam.ext.commands.Bot` you can do with :class:`Bot`.
    """



from __future__ import annotations

import re
from contextvars import ContextVar
from typing import TYPE_CHECKING

from betterproto.casing import pascal_case
from typing_extensions import TypeVar

from ...trade import Inventory, Item
from ...utils import cached_slot_property
from .currency import Metal
from .enums import *
from .protobufs import base, econ, struct_messages

if TYPE_CHECKING:
    from collections.abc import Iterable

    from ...abc import BaseUser, PartialUser
    from ...user import User
    from .client import ClientUser
    from .state import GCState
    from .types.schema import Schema

__all__ = (
    "BackpackItem",
    "Backpack",
)


WEAR_PARSER = re.compile("|".join(re.escape(wear.name) for wear in WearLevel))
SCHEMA = ContextVar["Schema"]("SCHEMA")
SKU_RE = re.compile(
    r"""\

""",
    re.VERBOSE,
)


OwnerT = TypeVar("OwnerT", bound="PartialUser", default="BaseUser", covariant=True)


class BackpackItem(Item[OwnerT]):
    """A class to represent an item from the client's backpack.

    Note
    ----
    This is meant to be user instantiable but only to use the following methods:

        - :meth:`is_australium`
        - :meth:`is_craftable`
        - :meth:`is_unusual`
        - :attr:`wear`
        - :attr:`equipable_by`
        - :attr:`slot`
    """

    __slots__ = (
        "position",
        "account_id",
        "inventory",
        "quantity",
        "level",
        "flags",
        "origin",
        "custom_name",
        "custom_description",
        "attribute",
        "interior_item",
        "in_use",
        "style",
        "original_id",
        "contains_equipped_state",
        "equipped_state",
        "contains_equipped_state_v2",
        "_quality_cs",
        "_def_index_cs",
    )
    REPR_ATTRS = (*Item.REPR_ATTRS, "position", "def_index")
    _state: GCState

    position: int
    """The item's position in the backpack."""

    account_id: int
    """Same as the :attr:`steam.ID.id` of the :attr:`steam.Client.user`."""
    inventory: int
    """The attribute the :attr:`position` is calculated from."""
    quantity: int
    """I think this should be the same as :attr:`amount`."""
    level: int
    """The item's level."""
    flags: ItemFlags
    """The item's flags."""
    origin: ItemOrigin
    """The item's origin."""
    custom_name: str
    custom_description: str
    attribute: list[base.ItemAttribute]
    interior_item: base.Item
    in_use: bool
    style: int
    original_id: int
    contains_equipped_state: bool
    equipped_state: list[base.ItemEquipped]
    contains_equipped_state_v2: bool

    # the other attribute definitions others not a clue please feel free to PR them

    @cached_slot_property
    def quality(self) -> ItemQuality:
        """The item's quality."""
        for tag in self.tags:
            if tag.category == "Quality":
                try:
                    return ItemQuality[tag.internal_name.title().replace(" ", "")]
                except KeyError:
                    pass
        raise RuntimeError(f"could not find quality for {self}")

    async def use(self: BackpackItem[ClientUser]) -> None:
        """Use this item."""
        await self._state.ws.send_gc_message(base.UseItem(item_id=self.id))

    async def open(self: BackpackItem[ClientUser], key: BackpackItem) -> None:
        """Open a crate with a ``key``.

        Parameters
        ----------
        key
            The key to open the crate with.
        """
        await self._state.ws.send_gc_message(struct_messages.OpenCrateRequest(key_id=key.id, crate_id=self.id))

    async def delete(self: BackpackItem[ClientUser]) -> None:
        """Delete this item."""
        await self._state.ws.send_gc_message(struct_messages.DeleteItemRequest(item_id=self.id))

    async def wrap(self: BackpackItem[ClientUser], wrapper: BackpackItem[ClientUser]) -> None:
        """Wrap this item with the ``wrapper``.

        Parameters
        ----------
        wrapper
            The wrapping paper to use.
        """
        await self._state.ws.send_gc_message(
            struct_messages.WrapItemRequest(item_id=self.id, wrapping_paper_id=wrapper.id)
        )

    async def unwrap(self: BackpackItem[ClientUser]) -> None:
        """Unwrap this item."""
        await self._state.ws.send_gc_message(struct_messages.UnwrapItemRequest(gift_id=self.id))

    async def equip(self: BackpackItem[ClientUser], mercenary: Mercenary, slot: ItemSlot) -> None:
        """Equip this item to a mercenary.

        Parameters
        ----------
        mercenary
            The mercenary to equip the item to.
        slot
            The item slot to equip the item to.
        """
        if slot >= ItemSlot.Misc:
            raise ValueError("cannot use enums that aren't real item slots")
        await self._state.ws.send_gc_message(
            base.AdjustItemEquippedState(item_id=self.id, new_class=mercenary, new_slot=slot)
        )

    async def set_position(self: BackpackItem[ClientUser], position: int) -> None:
        """Set the position for this item.

        Parameters
        ----------
        position
            The position to set the item to. This is 0 indexed.
        """
        assert self._state.backpack is not None
        await self._state.backpack.set_positions([(self, position)])

    async def set_style(self: BackpackItem[ClientUser], style: int) -> None:
        """Set the style for this item."""
        await self._state.ws.send_gc_message(struct_messages.SetItemStyleRequest(item_id=self.id, style=style))

    async def send_to(self: BackpackItem[ClientUser], user: User) -> None:
        """Send this gift-wrapped item to another user.

        Parameters
        ----------
        user
            The user to send this gift wrapped item to.
        """
        await self._state.ws.send_gc_message(struct_messages.DeliverGiftRequest(user_id64=user.id64, gift_id=self.id))

    # async def remove_attribute(self, attribute: Attribute) -> None:
    #     await self._state.ws.remove_attribute

    # methods similar to https://github.com/danocmx/node-tf2-item-format

    def is_australium(self) -> bool:
        """Whether or not the item is australium."""
        return "Australium" in self.name and self.name != "Australium Gold"

    def is_craftable(self) -> bool:
        """Whether or not the item is craftable."""
        return all(description.value != "( Not Usable in Crafting )" for description in self.descriptions)

    def is_festivized(self) -> bool:
        """Whether or not the item is festivized."""
        return any(
            description.value == "Festivized" and description.color == "ffd700" for description in self.descriptions
        )

    def is_festive(self) -> bool:
        """Whether or not the item is festive."""
        return self.name.startswith("Festive ")

    @property
    def wear(self) -> WearLevel | None:
        """The item's wear level."""
        wear = WEAR_PARSER.findall(self.name)
        return WearLevel[wear[0]] if wear else None

    @property
    def equipable_by(self) -> list[Mercenary]:
        """The mercenaries the item is equipable."""
        tags = [tag for tag in self.tags if tag.category == "Class"]
        return [Mercenary[mercenary.internal_name] for mercenary in tags]

    @property
    def slot(self) -> ItemSlot | None:
        """The item's equip slot."""
        for tag in self.tags:
            if tag.category == "Type" and tag.internal_name:
                try:
                    return ItemSlot[
                        pascal_case(tag.internal_name, strict=False)
                        .replace("Pda", "PDA")
                        .replace("Tf_Gift", "Gift")
                        .replace("Craft_Item", "CraftItem")
                    ]
                except KeyError:
                    return ItemSlot._new_member(name=tag.internal_name, value=-1)

    @cached_slot_property
    def def_index(self) -> int:
        """The item's def index. This is used to form the item's SKU."""
        for def_index, item in SCHEMA.get()["items"].items():
            if item.get("name") == self.name:
                return int(def_index)
        raise RuntimeError(f"Could not find def_index for {self.name}")

    @property
    def effect(self) -> str | None:
        """The item's unusual effect."""
        if self.quality != ItemQuality.Unusual:
            return
        for description in self.descriptions:
            if description.value.startswith("★ Unusual Effect: ") and description.color == "ffd700":
                return description.value.removeprefix("★ Unusual Effect: ")
        raise RuntimeError(f"Could not find unusual effect for {self.name}")

    @property
    def texture(self) -> int | None:
        """The item's texture."""
        for tag in self.tags:
            if tag.category == "texture":
                return int(tag.internal_name)

    @property
    def spells(self) -> list[str]:
        raise NotImplementedError

    @property
    def paint_colour(self) -> int | None:
        raise NotImplementedError

    @property
    def sku(self) -> str:
        """The item's SKU."""
        raise NotImplementedError
        parts = [str(self.def_index), ";", self.quality.value]

        if self.effect:
            parts.append(f";u{self.effect}")

        if self.is_australium():
            parts.append(";australium")

        if not self.is_craftable():
            parts.append(";uncraftable")

        if self.wear:
            parts.append(f";w{self.wear.value}")  # TODO check

        if self.texture:
            parts.append(f";pk{self.texture}")

        if self.elevated:
            parts.append(";strange")

        if self.killstreak:
            parts.append(f";kt-{self.killstreak}")

        if self.def_index:
            parts.append(f";td-{self.def_index}")

        if self.festivized:
            parts.append(";festive")

        if self.craft_number:
            parts.append(f";n{self.selfNumber.value}")

        if self.crate_number:
            parts.append(f";c{self.selfNumber.value}")

        return "".join(parts)

    @classmethod
    def from_sku(cls, value: str, /) -> BackpackItem[None]:  # type: ignore  # this is cursed, should probably be changed
        """Construct a :class:`BackPackItem` from an SKU.

        Parameters
        ----------
        value
            The SKU to construct the item from.

        Raises
        ------
        ValueError
            The SKU is invalid.
        """
        raise NotImplementedError
        self = cls.__new__(cls)

        data = SKU_RE.match(value)
        if data is None:
            raise ValueError(f"Invalid SKU: {value!r}")
        self.def_index = int(data["def_index"])

        return self

    # - to_listing?
    # - from_listing?
    # https://github.com/ZeusJunior/node-tf2-backpack/blob/7ce6fd0fc2ec648cee239c276742af2d3ab3fe5c/src/parser.ts#L19


class Backpack(Inventory[BackpackItem["ClientUser"], "ClientUser"]):
    """A class to represent the client's backpack."""

    __slots__ = ()

    @property
    def metal(self) -> Metal:
        """Returns the amount of metal in this inventory"""
        return Metal.from_items(self)

    async def set_positions(self, items_and_positions: Iterable[tuple[BackpackItem[ClientUser], int]]) -> None:
        """Set the positions of items in the inventory.

        Parameters
        ----------
        items_and_positions
            A list of (item, position) pairs to set the positions for. This is 0 indexed.
        """
        await self._state.ws.send_gc_message(
            base.SetItemPositions(
                item_positions=[
                    base.SetItemPositionsItemPosition(item_id=item.id, position=position)
                    for item, position in items_and_positions
                ],
            )
        )

    async def sort(self, type: BackpackSortType) -> None:
        """Sort this inventory.

        Parameters
        ----------
        type
            The sort type to sort by, only types visible in game are usable.
        """
        await self._state.ws.send_gc_message(base.SortItems(sort_type=type))

    async def trade_up(
        self, *items: BackpackItem[ClientUser]
    ) -> None:  # TODO would be nice to have a way to be able to return the new item
        """Trade up these items. Into a new item.

        Note
        ----
        The `TF2 wiki <https://wiki.teamfortress.com/wiki/Trade-Up#Item_Grade_Trade-Up>`_ for more info.
        """
        await self._state.ws.send_gc_message(econ.CraftCollectionUpgrade([item.id for item in items]))



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from ipaddress import IPv4Address
from typing import TYPE_CHECKING, Any, Generic, TypeAlias, TypeVar

from typing_extensions import TypedDict
from yarl import URL

if TYPE_CHECKING:
    from collections.abc import Coroutine

T = TypeVar("T")


class ResponseDict(TypedDict, Generic[T]):
    response: T


Coro: TypeAlias = "Coroutine[Any, Any, T]"
ResponseType: TypeAlias = "Coro[ResponseDict[T]]"
StrOrURL: TypeAlias = URL | str
IPAdress: TypeAlias = IPv4Address


class CMList(TypedDict):
    success: bool
    serverlist: list[CM]


class CM(TypedDict):
    endpoint: str
    load: float
    wtd_load: float


class EResultSuccess(TypedDict):
    success: int


class AddWalletCode(EResultSuccess):
    amount: int
    detail: int



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Final, cast

from .abc import PartialUser
from .app import PartialApp
from .clan import PartialClan
from .enums import EventType, Type, UserNewsType
from .id import ID
from .package import PartialPackage
from .types.id import PostID, PublishedFileID
from .utils import DateTime

if TYPE_CHECKING:
    from .achievement import UserNewsAchievement
    from .event import Announcement, Event
    from .friend import Friend
    from .post import Post
    from .protobufs import user_news
    from .published_file import PublishedFile
    from .review import Review
    from .state import ConnectionState


class UserNews:
    """Represents a news entry on the activity feed."""

    def __init__(
        self,
        state: ConnectionState,
        news: user_news.GetUserNewsResponse.Event,
        achievements: dict[str, UserNewsAchievement],
    ) -> None:
        self._state = state
        self.type: Final = UserNewsType.try_value(news.eventtype)
        """The type of news entry."""
        self.created_at = DateTime.from_timestamp(news.eventtime)
        """When the news entry was created."""
        target = ID(news.steamid_target)
        actor = ID(news.steamid_actor)
        self.target = cast(
            "Friend | PartialUser | PartialClan | None",
            (
                (state._maybe_friend(target.id) if target.type is Type.Individual else PartialClan(state, target.id64))
                if news.steamid_target
                else None
            ),
        )
        """The target of the news entry."""
        self.actor = cast(
            "Friend | PartialUser | PartialClan",
            state._maybe_friend(actor.id) if actor.type is Type.Individual else PartialClan(state, actor.id64),
        )
        """The actor of the news entry."""
        self.app = PartialApp(state, id=news.gameid) if news.gameid else None
        """The app this news entry is for."""
        self.package = PartialPackage(state, id=news.gameid) if news.gameid else None
        """The package this news entry is for."""
        self.apps = [PartialApp(state, id=id) for id in news.appids]
        """If :attr:`type` is :attr:`UserNewsType.ReceivedNewGame`, the apps unlocked by :attr:`package`.
        If :attr:`type` is :attr:`UserNewsType.AddedGameToWishlist`, the apps wishlisted by :attr:`actor`.
        """
        self.achievements = [achievements[name] for name in news.achievement_names]
        """The achievements unlocked. Only used for :attr:`UserNewsType.AchievementUnlocked`."""
        self.shortcut_id = news.shortcutid
        """The shortcut ID this news entry is for."""
        self.event_id = news.clan_eventid
        """The event ID this news entry is for. Only used for :attr:`UserNewsType.ScheduledEvent`."""
        self.announcement_id = news.clan_announcementid
        """The announcement ID this news entry is for. Only used for :attr:`UserNewsType.PostedAnnouncement`."""
        self.published_file_id = PublishedFileID(news.publishedfileid)
        """The published file ID this news entry is for."""
        self.post_id = PostID(news.steamid_target if self.type is UserNewsType.Post else 0)  # lol valve
        """The post ID this news entry is for. Only used for :attr:`UserNewsType.Post`."""
        if self.post_id:
            self.target = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} type={self.type!r} target={self.target!r} actor={self.actor!r}>"

    # ideally these would use self: UserNews[Literal[Type.Event]] etc but that doesn't work with pyright unfortunately
    # see microsoft/pyright#5706
    async def event(self) -> Event[EventType, PartialClan]:
        """Fetch the event this news entry is for."""
        if not self.event_id:
            raise ValueError("There's no associated Event for this news entry")
        assert isinstance(self.actor, PartialClan)
        return await self.actor.fetch_event(self.event_id)

    async def announcement(self) -> Announcement[PartialClan]:
        """Fetch the announcement this news entry is for."""
        if not self.announcement_id:
            raise ValueError("There's no associated Announcement for this news entry")
        assert isinstance(self.actor, PartialClan)
        return await self.actor.fetch_announcement(self.announcement_id)

    async def published_file(self) -> PublishedFile:
        """Fetch the published file this news entry is for."""
        if not self.published_file_id:
            raise ValueError("There's no associated PublishedFile for this news entry")
        assert isinstance(self.actor, PartialUser)
        file = await self.actor.fetch_published_file(self.published_file_id)
        assert file
        return file

    async def review(self) -> Review:
        """Fetch the review this news entry is for."""
        if self.type is not UserNewsType.Review:
            raise ValueError("There's no associated Review for this news entry")
        assert isinstance(self.actor, PartialUser)
        assert self.app
        return await self.actor.fetch_review(self.app)

    async def post(self) -> Post:
        """Fetch the post this news entry is for."""
        if self.type is not UserNewsType.Post:
            raise ValueError("There's no associated Post for this news entry")
        assert isinstance(self.actor, PartialUser)
        return await self.actor.fetch_post(self.post_id)

    # async def comment(self) -> Comment:
    #     """Fetch the comment this news entry is for."""  # TODO not sure how this works yet
    #     if self.type is UserNewsType.CommentByMe:
    #         return await self.actor.fetch_comment(self.target)
    #     elif self.type is UserNewsType.CommentOnMe:
    #         return await self._state.user.fetch_comment(self.target)
    #     raise ValueError("There's no associated Comment for this news entry")
